{-# LANGUAGE DeriveDataTypeable, DeriveGeneric, DefaultSignatures #-}

{-|
  Implements the basic concepts of combinatory categorial grammar: items, grammars,
  syntactic categories, and combinators. A small set of functions to work with these is also
  provided.
  
  In addition, some amount of simple statistics tracking exists.
-}
module ELCS.CCG.CCG (
-- * CCG basics
  CCGItem(..)
, Grammar
, SynCat(..)
, Combinator(..)
, IndexedGrammar

-- * Dealing with combinators
, CombinatorDirection(..)
, combinatorDirection

-- * Dealing with syntactic categories
, formatSynCat
, resultingSynCat

-- * Statistics
, StatsGrammar
, ItemStats(..)
, initialStats
, combineStats

-- * References
-- $references
) where

import qualified Data.Map as Map
import Control.DeepSeq.Generics
import GHC.Generics(Generic)

-- CCG basics

-- | A CCG item (a.k.a. lexical item) is the basic element of a CCG grammar. An item is also
--   produced at each intermediate step of a CCG derivation. A CCG item combines a contiguous
--   span of tokens (i.e. words from the natural language sentence), a syntactic category,
--   and a semantic category.
--
--   The type of semantics used is left unspecified, but implementations of semantic formalisms
--   will generally be instances of 'ELCS.MR.MR', as is required by most of the algorithms used
--   here.
data CCGItem sem = CCGItem { ccgTokens :: [String]
                           , ccgSyntax :: SynCat
                           , ccgSemantics :: sem }
                   deriving (Eq, Show, Ord, Read, Generic)

instance (NFData sem) => NFData (CCGItem sem) where rnf = genericRnf

-- | A grammar is simply a list of CCG items.
type Grammar a = [CCGItem a]

-- | Defines the set of syntactic categories.
--
--   Following [Kwiatkowski et al. (2011)], we introduce the non-standard vertical slash which
--   can be applied in both directions.
--   The vertical slash is generated when deducing the syntactic type from the
--   semantic type (which does not specify direction).
--
--   In the slash types, the result type is always the left component.
data SynCat = Atom String (Maybe String)  -- ^ An atom with atomic category and an optional feature
            | ForwardSlash SynCat SynCat  -- ^ Forward slash
            | BackwardSlash SynCat SynCat -- ^ Backward slash
            | VerticalSlash SynCat SynCat -- ^ Vertical (i.e. non-directional) slash
            deriving (Show, Eq, Ord, Read, Generic)

instance NFData SynCat where rnf = genericRnf
  
-- | Defines the set of CCG combinators.
data Combinator = FA -- ^ Forward application.
                | BA -- ^ Backward application.
                deriving (Eq, Show, Ord, Read, Generic)
                
instance NFData Combinator where rnf = genericRnf


-- | A grammar where items are indexed by tokens. 
type IndexedGrammar a = Map.Map [String] [CCGItem a]


-- Dealing with combinators

-- | Data type for specifying the directionality of combinators, i.e. the respective positions
--   of head and complement.
data CombinatorDirection = Forward  -- ^ Indicates that the head is on the left side.
                         | Backward -- ^ Indicates that the head is on the right side.
                         deriving (Eq,Show)

-- | Returns the directionality of a given combinator.
combinatorDirection :: Combinator -> CombinatorDirection
combinatorDirection c =
  case c of
    FA -> Forward
    BA -> Backward

  
-- Dealing with syntactic categories
  
-- | Pretty-prints a syntactic category in standard CCG notation.
formatSynCat :: SynCat -> String
formatSynCat c = case c of
  Atom n Nothing -> n
  Atom n (Just f) -> n ++ "[" ++ f ++ "]"
  ForwardSlash c1 c2 -> "(" ++ (formatSynCat c1) ++ "/" ++ (formatSynCat c2) ++ ")"
  BackwardSlash c1 c2 -> "(" ++ (formatSynCat c1) ++ "\\" ++ (formatSynCat c2) ++ ")"
  VerticalSlash c1 c2 -> "(" ++ (formatSynCat c1) ++ "|" ++ (formatSynCat c2) ++ ")"
  
-- | Computes the syntactic category, if any, resulting from a combinator application.
resultingSynCat :: Combinator -> (SynCat,SynCat) -> Maybe SynCat
resultingSynCat comb (cat1,cat2) = case comb of
  FA -> case cat1 of
          ForwardSlash r t -> if cat2 == t then Just r else Nothing
          _                -> Nothing
  BA -> case cat2 of
          BackwardSlash r t -> if cat1 == t then Just r else Nothing
          _                 -> Nothing


-- Statistics

-- | A grammar with annotated statistics.
type StatsGrammar a = [(CCGItem a, ItemStats)]

-- | Statistics that may be annotated on an item.
data ItemStats =
  ItemStats
  { statSentenceCount :: Int -- ^ The number of sentences this item occurs in
  } deriving (Eq, Show, Ord, Generic)

instance NFData ItemStats where rnf = genericRnf

-- | Statistics for a newly created item.
initialStats :: ItemStats
initialStats =
  ItemStats
  { statSentenceCount = 1
  }

-- | Combines the statistics of two occurrences of the same item.
combineStats :: ItemStats -> ItemStats -> ItemStats
combineStats s1 s2 = ItemStats
  { statSentenceCount = statSentenceCount s1 + statSentenceCount s2
  }
  
          
{-$references
  * [Kwiatkowski et al. (2011)] Kwiatkowski, Tom, Luke Zettlemoyer, Sharon Goldwater, and Mark Steedman. ‘Lexical Generalization in CCG Grammar Induction for Semantic Parsing’. In Proceedings of the Conference on Empirical Methods in Natural Language Processing, 1512–1523, 2011. <http://dl.acm.org/citation.cfm?id=2145593 PDF@ACM>.
-}
