{-# LANGUAGE DeriveGeneric #-}

module ELCS.CCG.Derivation (
  -- * Derivations
  Derivation
, NodeData
, NodeType(..)
, Token
, derivChildren
, derivNodeType
, derivTokens
, derivSynCat
, derivSemantics
, mkDerivation

  -- * Extracting syntax trees
, toSyntaxTree
, toPrettySyntaxTree

  -- * Extracting CCG items
, rootItem

  -- * Sanity checking
, SanityError(..)
, isSane
, saneDerivationOrError
) where

import ELCS.LambdaLike
import ELCS.MR
import ELCS.MR.Label
import ELCS.CCG.CCG
import ELCS.MR.Type
import ELCS.DRT hiding(mkAbs,mkApp,mkVExp,isSane)
import ELCS.Util

import Data.Maybe
import Data.List
import Data.Tree
import Control.Exception
import Control.DeepSeq.Generics
import GHC.Generics(Generic)
--import Debug.Trace


-- Derivations

-- | Derivations are represented using 'Data.Tree'.
--   Each node is annotated with two pieces of information:
--
--     * a node type which specifies the kind of derivation rule represented by the node
--
--     * a 'CCGItem' which corresponds to the tokens, syntactic type and meaning spanned by the
--       derivation node.
type Derivation = Tree NodeData

-- | Annotation for 'Derivation' nodes.
data NodeData = NodeData
  { nodeType  :: NodeType     -- ^ Describes the semantics of the derivation node.
  , nodeItem  :: CCGItem DRT  -- ^ The semantics saved at this node
  } deriving (Show, Eq, Generic)

instance NFData NodeData where rnf = genericRnf

-- | Describes a derivation rule.
data NodeType =
    RuleNode String -- ^ A node where a rule is applied. The string is one of the combinator types
                    --   used in the GMB.
  | LexNode         -- ^ A lexical item node. These are leaves of the derivation.
  deriving (Show, Eq, Generic)

instance NFData NodeType where rnf = genericRnf

-- | Tokens are just strings.
type Token = String

derivChildren :: Derivation -> [Derivation]
derivChildren = subForest

derivNodeType :: Derivation -> NodeType
derivNodeType = nodeType . rootLabel

derivTokens :: Derivation -> [Token]
derivTokens = ccgTokens . nodeItem . rootLabel

derivSynCat :: Derivation -> SynCat
derivSynCat = ccgSyntax . nodeItem . rootLabel

derivSemantics :: Derivation -> DRT
derivSemantics = ccgSemantics . nodeItem . rootLabel

mkDerivation :: NodeType -> CCGItem DRT -> [Derivation] -> Derivation
mkDerivation nt item children = let
    nodeData = NodeData nt item
  in Node nodeData children


-- Extraction of syntax trees

-- | Extracts a syntax tree annotated with the spans for each node.
toSyntaxTree :: Derivation -> Tree (Int,Int)
toSyntaxTree d = unfoldTree unfoldSyntaxTree (d,tokenSpan)
  where tokenSpan = (0, (length $ derivTokens d) - 1)

unfoldSyntaxTree :: (Derivation,(Int,Int)) -> ((Int,Int), [(Derivation,(Int,Int))])
unfoldSyntaxTree (d,sp) = (sp, children)
  where
    children = annotateSpans sp (derivChildren d)

annotateSpans :: (Int,Int) -> [Derivation] -> [(Derivation,(Int,Int))]
annotateSpans sp ds = snd $ foldr annotateSpan (sp,[]) ds

annotateSpan :: Derivation
             -> ((Int,Int),[(Derivation,(Int,Int))])
             -> ((Int,Int),[(Derivation,(Int,Int))])
annotateSpan d ((a,b),anns) = let
    b'  = b - (length $ derivTokens d)
    ann = (d,(b'+1,b))
  in ((a,b'),ann:anns)

-- | Extracts a syntax tree, with rule names at each node and words at the leaves.
toPrettySyntaxTree :: Derivation -> Tree String
toPrettySyntaxTree = unfoldTree unfold
  where
    unfold :: Derivation -> (String, [Derivation])
    unfold d = (nodeLabel d, derivChildren d)

    nodeLabel d = case derivNodeType d of
                    RuleNode s -> s
                    LexNode -> concat $ intersperse "," $ derivTokens d
                               -- We should actually have only one token per lex node,
                               -- but we handle the more general case.


-- Extracting CCG items

-- | Returns the sentential root item of a derivation.
rootItem :: Derivation -> Either String (CCGItem DRT)
rootItem d = do
    let sem = makeVariablesUnique $ normalForm $
              removeInnerAlignments $ alignedSemantics d
    semTyped <- case inferTypes sem of
                  Right s' | isCompleteType (fromJust $ drtType s') -> Right s'
                           | otherwise -> Left "rootItem: Incomplete type"
                  Left e -> Left ("rootItem: Type error (" ++ e ++ ")")
    semNormalized <- case normalizeDRT semTyped of
                       Right s' -> Right s'
                       Left err -> Left (show err)
    return $ CCGItem { ccgTokens = derivTokens d,
                       ccgSyntax = Atom "S" Nothing,
                       ccgSemantics = semNormalized }


-- Sanity checking

data SanityError = WrongChildCount
                 | UnknownRuleType String
                 | SyntaxError SyntaxError
                 | UnexpectedFreeVars
                 | VarsNotUnique
                 deriving (Show)

-- | Sanity checking of a derivation.
--   Returns a list of all the things that are wrong with the derivation.
sanityCheck :: Derivation -> [(Derivation, SanityError)]
sanityCheck d = let
    errs = concat $
           map ELCS.CCG.Derivation.sanityCheck (derivChildren d) ++
           [checkNodeSane d] ++
           [checkSemantics d (alignedSemantics d)]
  in errs

saneDerivationOrError :: Derivation -> Either [(Derivation, SanityError)] Derivation
saneDerivationOrError d =
  let
    errors = ELCS.CCG.Derivation.sanityCheck d
  in if null errors then Right d else Left errors


isSane :: Derivation -> Bool
isSane d = null $ ELCS.CCG.Derivation.sanityCheck d

checkNodeSane :: Derivation -> [(Derivation, SanityError)]
checkNodeSane d = let
      errs = concat $
             [checkChildCount d] ++
             [checkSemantics d (derivSemantics d)]
  in errs

checkChildCount :: Derivation -> [(Derivation, SanityError)]
checkChildCount d =
  let
    childCount = length $ derivChildren d
    checkCount i = if childCount == i then [] else [(d, WrongChildCount)]
    result =
      case derivNodeType d of
        LexNode -> checkCount 0
        RuleNode "tc" -> checkCount 1
        RuleNode "ftr" -> checkCount 1
        RuleNode "btr" -> checkCount 1
        RuleNode "fa" -> checkCount 2
        RuleNode "ba" -> checkCount 2
        RuleNode "fc" -> checkCount 2
        RuleNode "bc" -> checkCount 2
        RuleNode "fxc" -> checkCount 2
        RuleNode "bxc" -> checkCount 2
        RuleNode "gfc" -> checkCount 2
        RuleNode "gbc" -> checkCount 2
        RuleNode "fbxc" -> checkCount 2
        RuleNode "gbxc" -> checkCount 2
        RuleNode "conj" -> checkCount 2
        RuleNode r -> [(d, UnknownRuleType r)]
  in result

checkSemantics :: Derivation -> DRT -> [(Derivation, SanityError)]
checkSemantics d sem = let
    errs = concat $
           [maybe [] (\x -> [(d, SyntaxError x)]) (ELCS.DRT.sanityCheck sem)] ++
           [(if null (freeVars sem) then [] else [(d, UnexpectedFreeVars)])]
           -- todo this clashes with some GMB lexical entries
           -- ++ [if checkVariablesUnique sem then [] else [(d, VarsNotUnique)]]
  in errs

-- | Gives the semantics of the derivation with alignments annotated to each node.
-- Every alignment will be a contiguous range which
-- corresponds to the tokens which evoked the node in the derivation. Token indices start at 0.
alignedSemantics :: Derivation -> DRT
alignedSemantics derivation =
    annotate (0,length (derivTokens derivation) - 1) derivation
  where
    -- Annotates d with l and annotates the children accordingly.
    annotate :: (Int,Int) -> Derivation -> DRT
    annotate (f,l) d = let
        -- The semantics at this node.
        nodeSemantics = drtSetAlignmentRecursive [f..l] $ derivSemantics d
        label' = drtLabel nodeSemantics
      in makeVariablesUnique $ normalForm $
         case derivNodeType d of
           LexNode    -> nodeSemantics
           RuleNode r -> applyRule r label' nodeSemantics (recurse (f,l) (derivChildren d))

    -- recurse breaks the list of children down into individual calls to annotate.
    recurse l es = let
        -- To get the spans for the children, we fold over them, collecting the spans created so far
        -- and keeping the end index of the current child.
        spans = fst $ foldr getSpan ([],(snd l)) es
        getSpan c (ss,cur) = let
            start = cur - (length (derivTokens c)) + 1
          in ((start,cur):ss, start - 1)
      in map (uncurry annotate) $ zip spans es


-- Derivation rules

applyRule :: String -> MRLabel -> DRT -> [DRT] -> DRT
applyRule t l base children =
  case t of
    "fa"   -> applyFA l children
    "ba"   -> applyBA l children
    "fc"   -> applyFC l children
    "bc"   -> applyBC l children
    "fxc"  -> applyFC l children
    "bxc"  -> applyBC l children
    "conj" -> applyConj l children
    "tc"   -> base
    "ftr"  -> base -- Type raising – ignoring for now
    "btr"  -> base
    "gfc"  -> base -- Composition of functions with multiple arguments.
                   -- I’m deferring implementation for now.
    "gbc"  -> base
    "gfxc" -> base
    "gbxc" -> base
    x     -> throw $ FatalException ("Unknown rule type: " ++ x)

applyFA :: MRLabel -> [DRT] -> DRT
applyFA l [f,a] = mkAppRenaming l f a
applyFA _ _     = throw $ FatalException "Wrong number of arguments for FA combinator."

applyBA :: MRLabel -> [DRT] -> DRT
applyBA l [a,f] = mkAppRenaming l f a
applyBA _ _     = throw $ FatalException "Wrong number of arguments for BA combinator."

applyComposition :: MRLabel -> DRT -> DRT -> DRT
applyComposition l f g = mkAbs l x (mkAppRenaming l f (mkAppRenaming l g (mkVExp l x)))
  where
    x = inventVar "v" (union (vars f) (vars g))

applyFC :: MRLabel -> [DRT] -> DRT
applyFC l [f,g] = applyComposition l f g
applyFC _ _     = throw $ FatalException "Wrong number of arguments for FC combinator."

applyBC :: MRLabel -> [DRT] -> DRT
applyBC l [g,f] = applyComposition l f g
applyBC _ _     = throw $ FatalException "Wrong number of arguments for BC combinator."

applyConj :: MRLabel -> [DRT] -> DRT
applyConj l [f,a] = mkAppRenaming l f a
applyConj _ _     = throw $ FatalException "Wrong number of arguments for Comp combinator."
