{-| Contains functions for loading and saving whole grammars, and other
    associated functions.

    Grammars are saved as JSON files, with every CCG item represented as a JSON object on a
    single line. The precise format is defined in 'ELCS.CCG.Serialise.JSON'.
    
    Grammars are also optionally GZip-compressed, and this module provides functionality to deal
    with this.
-}
module ELCS.CCG.LoadSave (
  SerialisationOptions(..)
, defaultSerialisationOptions
, CompressionSetting(..)
, Codec(..)

, processCompressByExt

-- Save
, hSaveList
, saveGrammar
, saveGrammarWith
, saveListWith

-- Load
, loadGrammar
, loadGrammarDRT
, loadGrammarWith
, loadListWith

-- Stats grammars
, loadStatsGrammarWith
, discardStats
, addInitialStats

-- Indexing
, indexGrammar

-- Save to HTML
, saveGrammarToHTML
, saveGrammarToHTMLChunked

-- Streaming operations on grammars
, makeMultiSentenceHist
, withStatsGrammar
, withList
) where

import ELCS.CCG.CCG
import ELCS.DRT
import ELCS.Util
import qualified ELCS.CCG.Serialise.JSON as SerJSON

import Data.List
import qualified Data.Map as Map
import Text.XML.HXT.Core
import Control.Monad
import Control.Applicative
import System.FilePath
import qualified Codec.Compression.GZip as GZip
import System.IO
import qualified Data.ByteString.Lazy as BS
import qualified Data.Aeson as Aes
--import Control.Exception
--import Debug.Trace


-- Saving and loading

data SerialisationOptions = SerialisationOptions
  { serCompress :: CompressionSetting
  , serCodec :: Codec
  , serWithStats :: Bool
  }
  deriving (Show, Eq)
  
defaultSerialisationOptions :: SerialisationOptions
defaultSerialisationOptions = SerialisationOptions
  { serCompress = CompressByExt
  , serCodec = JSON
  , serWithStats = False }
  
data CompressionSetting = Compressed | Uncompressed | CompressByExt
     deriving (Show, Eq)
     
data Codec = JSON
     deriving (Show, Eq)

-- | Saves a list to a handle, optionally applying gzip compression.
hSaveList :: (Aes.ToJSON a)
          => SerialisationOptions
          -> [a]       -- ^ The list to save.
          -> Handle    -- ^ The handle to save to.
          -> IO ()
hSaveList opts g h = do
  let c = case serCodec opts of
            JSON -> SerJSON.encode g
  let c' = if (serCompress opts == Compressed) then GZip.compress c else c
  BS.hPut h c'

-- | Saves a grammar to a file with default options.
saveGrammar :: (Aes.ToJSON a) => FilePath -> Grammar a -> IO ()
saveGrammar = saveGrammarWith defaultSerialisationOptions
  
-- | Saves a grammar to a file, optionally applying gzip compression. Compression is recommended,
--   as grammars have a high level of redundancy and compress to about 1–5% of their original size.
saveGrammarWith :: (Aes.ToJSON a)
                => SerialisationOptions
                -> FilePath  -- ^ The path to save to. ".gz" is appended if compression is enabled.
                -> Grammar a -- ^ The grammar to save.
                -> IO () -- ^ The file that was actually written to.
saveGrammarWith = saveListWith
  
saveListWith :: (Aes.ToJSON a)
             => SerialisationOptions
             -> FilePath
             -> [a]
             -> IO ()
saveListWith opts p g = do
  let opts' = processCompressByExt opts p
  withBinaryFile p WriteMode (hSaveList opts' g)

  
-- | Determines whether the path is to a compressed file.
processCompressByExt :: SerialisationOptions
                     -> FilePath -- ^ The original file path.
                     -> SerialisationOptions
processCompressByExt opts p =
  case serCompress opts of
    CompressByExt ->
      if takeExtension p == ".gz"
        then opts { serCompress = Compressed }
        else opts { serCompress = Uncompressed }
    _ -> opts

-- | Loads a DRT grammar from a file. This is the same as 'loadGrammar', just with a more specific
--   type.
loadGrammarDRT :: FilePath -> IO (Grammar DRT)
loadGrammarDRT = loadGrammar

-- | Loads a grammar from a file.
--   If the file name ends in ".gz", gzip decompression is applied.
loadGrammar :: (Aes.FromJSON a) => FilePath -> IO (Grammar a)
loadGrammar = loadGrammarWith defaultSerialisationOptions

loadGrammarWith :: (Aes.FromJSON a) => SerialisationOptions -> FilePath -> IO (Grammar a)
loadGrammarWith = loadListWith

loadListWith :: (Aes.FromJSON a) => SerialisationOptions -> FilePath -> IO [a]
loadListWith opts fPath = do
  contents <- readFileWithOptions opts fPath
  return $ case serCodec opts of
             JSON -> SerJSON.decode contents
  
readFileWithOptions :: SerialisationOptions -> FilePath -> IO (BS.ByteString)
readFileWithOptions opts p = do
  let compressed = case serCompress opts of
                     CompressByExt -> takeExtension p == ".gz"
                     Compressed    -> True
                     Uncompressed  -> False
  contents <- BS.readFile p
  let contents' = if compressed then GZip.decompress contents else contents
  return contents'


-- Stats grammars

loadStatsGrammarWith :: (Aes.FromJSON a) => SerialisationOptions -> FilePath -> IO (StatsGrammar a)
loadStatsGrammarWith = loadListWith

discardStats :: StatsGrammar a -> Grammar a
discardStats = map fst

addInitialStats :: Grammar a -> StatsGrammar a
addInitialStats = map (\x -> (x, initialStats))
  
  
-- Indexing

-- | Indexes a grammar by surface strings.
indexGrammar :: (Ord a) => Grammar a -> IndexedGrammar a
indexGrammar g = bucketBy ccgTokens g

bucketBy :: (Ord a, Ord b) => (a -> b) -> [a] -> Map.Map b [a]
bucketBy f s = foldl' (addToBucket f) (Map.empty) s

addToBucket :: (Ord a, Ord b) => (a -> b) -> Map.Map b [a] -> a -> Map.Map b [a]
addToBucket f m e = let
    bName = f e
    bucket = Map.lookup bName m
    bucket' = case bucket of
                Just b -> e:b
                Nothing -> [e]
  in Map.insert bName bucket' m
  

-- Saving to HTML

-- | Saves an HTML representation of the grammar.
saveGrammarToHTML :: FilePath -> Grammar DRT -> IO ()
saveGrammarToHTML f g = runWithInput (fromLA arrow >>> writeHtmlFile f) g
  where
    arrow :: LA (Grammar DRT) XmlTree
    arrow =
      root [] [
        selem "html" [
          selem "head" [
            aelem "link" [ attr "rel" (txt "stylesheet")
                         , attr "href" (txt "gmbstyle/style.css")
                         , attr "type" (txt "text/css")]
          , aelem "meta" [ attr "charset" (txt "UTF-8") ]
          ]
        , selem "body" [
            unlistA >>> writeItem
          ]
        ]
      ]
    writeItem =
      selem "p" [ showItemTokens >>> mkText ] <+>
      selem "p" [ showItemSyntax >>> mkText ] <+>
      selem "p" [ arr (ccgSemantics) >>> writeExp ]
    showItemTokens = arr (concat . (intersperse " ") . ccgTokens)
    showItemSyntax = arr (formatSynCat . ccgSyntax)

-- | Saves an HTML representation of the grammar, distributed over a number of documents so your
--   browser’s layout algorithm won’t choke on huge grammars.
saveGrammarToHTMLChunked :: Int       -- ^ The number of grammar items to include in each HTML file.
                         -> String    -- ^ A file name prefix. A ".x.html" suffix will be added,
                                      -- ^ where x is a running index.
                         -> Grammar DRT -- ^ The grammar to save.
                         -> IO ()
saveGrammarToHTMLChunked chunkSize f g = do
  let chunks = zip [(1::Int)..] (chunksOf chunkSize g)
  _ <- forM chunks $ \(i,c) ->
         saveGrammarToHTML (f ++ "." ++ show i ++ ".html") c
  return ()
  

-- Streaming operations on grammars

makeMultiSentenceHist :: StatsGrammar DRT -> Map.Map Int Int
makeMultiSentenceHist = (foldl' (incrMultiSentenceHist) Map.empty) . map snd

incrMultiSentenceHist :: Map.Map Int Int -> ItemStats -> Map.Map Int Int
incrMultiSentenceHist m i = let
    bucket = statSentenceCount i
  in Map.update (Just.(+1)) bucket m

withStatsGrammar :: FilePath -> (StatsGrammar DRT -> a) -> IO a
withStatsGrammar = withList

withList :: (Aes.FromJSON a) => FilePath -> ([a] -> b) -> IO b
withList p f = f <$> loadListWith defaultSerialisationOptions p
