{-# LANGUAGE FlexibleInstances #-}
module ELCS.CCG.Serialise.Collate (
  collateDRT
) where

import ELCS.CCG
import ELCS.DRT
import ELCS.Util

import qualified Data.List.Ordered as Ordered
import System.FilePath
import System.Directory
import System.IO
import Control.Exception
import Control.Applicative
import Debug.Trace
import qualified Data.Aeson as Aes


data InputFile = InputFile
  { ifPath :: FilePath
  , ifOptions :: SerialisationOptions
  , ifIsTemp :: Bool
  }
  deriving (Show, Eq)
  
  
class Collate a where
  load :: InputFile -> IO [a]
  hSave :: SerialisationOptions -> [a] -> Handle -> IO ()
  collate :: [a] -> [a] -> [a]
  
instance (Ord a, Aes.FromJSON a, Aes.ToJSON a) => Collate (CCGItem a) where
  load i = let
      opts = ifOptions i
    in if serWithStats opts
         then discardStats <$> loadStatsGrammarWith opts (ifPath i)
         else loadGrammarWith opts (ifPath i)
         
  hSave opts g h =
    if serWithStats opts
      then hSaveList opts (addInitialStats g) h
      else hSaveList opts g h

  collate = Ordered.union
  
instance (Ord a, Aes.FromJSON a, Aes.ToJSON a) => Collate (CCGItem a, ItemStats) where
  load i = let
      opts = ifOptions i
    in if serWithStats opts
         then loadStatsGrammarWith opts (ifPath i)
         else addInitialStats <$> loadGrammarWith opts (ifPath i)
         
  hSave opts g h =
    if serWithStats opts
      then hSaveList opts g h
      else hSaveList opts (discardStats g) h
      
  collate = loop
    where
      loop [] ys = ys
      loop xs [] = xs
      loop (x@(ix,sx):xs) (y@(_iy,sy):ys) =
        case compare x y of
          LT -> x : loop xs (y:ys)
          EQ -> (ix, combineStats sx sy) : loop xs ys
          GT -> y : loop (x:xs) ys
  

-- | Taking a list of grammar files, writes a new file that contains the union of all the grammars.
collateDRT :: SerialisationOptions -> SerialisationOptions -> [FilePath] -> FilePath -> IO ()
collateDRT inOpts outOpts is o = do
  -- First, we copy each input grammar to a tmp file
  let dir = takeDirectory o
  let tmpOpts = defaultSerialisationOptions
                  { serCompress = Compressed, serCodec = JSON, serWithStats = serWithStats outOpts }
  let ifs = map (\p -> InputFile p inOpts False) is
  final <- collateDRTBalanced tmpOpts dir ifs
  convert final (outOpts, o)
  if ifIsTemp final then removeFile (ifPath final) else return ()

collateDRTBalanced :: SerialisationOptions -> FilePath -> [InputFile] -> IO InputFile
collateDRTBalanced opts dir fs =
  case fs of
    [] -> throw $ FatalException "collateDRTBalanced: empty list"
    [x] -> return x
    [x,y] -> collateTwo opts dir x y
    _ -> do
           let l = ceiling $ (fromIntegral $ length fs) / (2::Float)
           x1 <- collateDRTBalanced opts dir (take l fs)
           x2 <- collateDRTBalanced opts dir (drop l fs)
           collateTwo opts dir x1 x2

collateTwo :: SerialisationOptions -> FilePath -> InputFile -> InputFile -> IO InputFile
collateTwo opts dir i1 i2 = do
  traceIO ("Collating " ++ (ifPath i1) ++ " with " ++ (ifPath i2))
  if serWithStats opts
    then collateTwoWithStats opts dir i1 i2
    else collateTwoNoStats opts dir i1 i2
    
collateTwoWithStats :: SerialisationOptions -> FilePath -> InputFile -> InputFile -> IO InputFile
collateTwoWithStats opts dir i1 i2 = do
  g1 <- load i1 :: IO (StatsGrammar DRT)
  g2 <- load i2 :: IO (StatsGrammar DRT)
  collateAndSave opts dir (i1,g1) (i2,g2)
  
collateTwoNoStats :: SerialisationOptions -> FilePath -> InputFile -> InputFile -> IO InputFile
collateTwoNoStats opts dir i1 i2 = do
  g1 <- load i1 :: IO (Grammar DRT)
  g2 <- load i2 :: IO (Grammar DRT)
  collateAndSave opts dir (i1,g1) (i2,g2)
  
collateAndSave :: (Aes.ToJSON a, Collate a) => SerialisationOptions -> FilePath
               -> (InputFile,[a]) -> (InputFile,[a]) -> IO InputFile
collateAndSave opts dir (i1,g1) (i2,g2) = do
  let g' = collate g1 g2
  o <- saveToTmp opts dir "temp-grammar" g'
  if ifIsTemp i1 then removeFile (ifPath i1) else return ()
  if ifIsTemp i2 then removeFile (ifPath i2) else return ()
  return $ InputFile o opts True

saveToTmp :: (Collate a, Aes.ToJSON a)
          => SerialisationOptions -> FilePath -> String -> [a]
          -> IO FilePath
saveToTmp opts dir template g = do
  (tmpFile,tmpH) <- openTempFile dir template
  hSave opts g tmpH
  hClose tmpH
  return tmpFile  

save :: (Collate a, Aes.ToJSON a)
     => SerialisationOptions -> FilePath -> [a] -> IO ()
save opts p g = do
  let opts' = processCompressByExt opts p
  withBinaryFile p WriteMode (hSave opts' g)

convert :: InputFile -> (SerialisationOptions, FilePath) -> IO ()
convert i (outOpts, outPath) =
  if serWithStats (ifOptions i)
    then (load i :: IO (StatsGrammar DRT)) >>=
         save outOpts outPath
    else (load i :: IO (Grammar DRT)) >>=
         save outOpts outPath
