{-# LANGUAGE TemplateHaskell, OverloadedStrings #-}
{-# OPTIONS_GHC -fno-warn-orphans #-}
module ELCS.CCG.Serialise.JSON (
  encode
, decode
) where

import ELCS.CCG.CCG
import ELCS.CCG.Split
import ELCS.DRT
import ELCS.MR
import ELCS.MR.Label
import ELCS.MR.Type
import ELCS.Util

import qualified Data.ByteString.Lazy as BS
import Data.Aeson hiding(encode,decode)
import qualified Data.Aeson
import Data.Aeson.Types
import Data.Aeson.TH
import Data.Maybe
import Data.Char
import GHC.Word(Word8)
import Control.Applicative
import Control.Monad
import Control.Exception
import Data.Text(Text,unpack)

encode :: (ToJSON a) => [a] -> BS.ByteString
encode xs = BS.intercalate (BS.singleton delimiter) $
            map Data.Aeson.encode xs

decode :: (FromJSON a) => BS.ByteString -> [a]
decode c = map fromJust $ map Data.Aeson.decode $ BS.split delimiter c

delimiter :: Word8
delimiter = fromIntegral $ fromEnum '\n'


(==!) :: Parser Text -> Text -> Parser ()
(==!) pv t = do
  v <- pv
  if (v == t)
    then return ()
    else throw $ FatalException ("Found " ++ unpack v ++ " where " ++ unpack t ++ " was expected")
  
mkTag :: String -> [Pair]
mkTag s = [ "tag" .= s ]


-- CCG

$(deriveJSON defaultOptions{fieldLabelModifier = (map toLower) . (drop 3)} ''CCGItem)
$(deriveJSON defaultOptions ''Combinator)
$(deriveJSON defaultOptions ''Split)

-- SynCat

instance ToJSON SynCat where
  toJSON (Atom c mf) = let
      feat = case mf of
               Just f -> [ "feature" .= f ]
               Nothing -> []
    in object $ mkTag "Atom" ++ [ "cat" .= c ] ++ feat
  toJSON (ForwardSlash l r ) = slashToJSON "ForwardSlash" l r
  toJSON (BackwardSlash l r ) = slashToJSON "BackwardSlash" l r
  toJSON (VerticalSlash l r ) = slashToJSON "VerticalSlash" l r
  
slashToJSON :: String -> SynCat -> SynCat -> Value
slashToJSON t l r =
  object [ "tag" .= t, "left" .= l, "right" .= r ]
  
instance FromJSON SynCat where
  parseJSON (Object v) = do
    tag <- v .: "tag" :: Parser Text
    case tag of
      "Atom" -> parseAtom v
      "ForwardSlash" -> parseSlash ForwardSlash v
      "BackwardSlash" -> parseSlash BackwardSlash v
      "VerticalSlash" -> parseSlash VerticalSlash v
      _ -> throw $ FatalException ("Invalid tag for syntactic category: " ++ unpack tag)
  parseJSON _ = mzero
      
parseAtom :: Object -> Parser SynCat
parseAtom v = Atom <$> v .: "cat" <*> v .:? "feature"
  
parseSlash :: (SynCat -> SynCat -> SynCat) -> Object -> Parser SynCat
parseSlash slash v = slash <$> v .: "left" <*> v .: "right"


-- DRT

instance ToJSON DRT where
  toJSON (DRT l h s) = object $ (headToPairs h) ++ (labelToPairs l) ++ (subsToPairs h s)
  
labelToPairs :: MRLabel -> [Pair]
labelToPairs l = [ "type" .= mrLabelType l, "alignment" .= mrLabelAlignment l ]
  
headToPairs :: DRTHead -> [Pair]
headToPairs (Expr h) = exprHeadToPairs h
headToPairs (Cond h) = condHeadToPairs h
headToPairs (ELCS.DRT.Const h) = constHeadToPairs h

subsToPairs :: DRTHead -> [DRT] -> [Pair]
subsToPairs (Expr (DRSHead _))  s = [ "conditions" .= s ]
subsToPairs (Expr (SDRSHead _)) s = [ "constituents" .= s ]
subsToPairs _                  [] = []
subsToPairs _                   s = [ "sub-expressions" .= s ]

exprHeadToPairs :: ExprHead -> [Pair]
exprHeadToPairs (VExpHead v) =
  mkTag "variable" ++ [ "variable" .= formatVar v ]
exprHeadToPairs (AbsHead v) =
  mkTag "lambda" ++ [ "variable" .= formatVar v ]
exprHeadToPairs AppHead =
  mkTag "application"
exprHeadToPairs (DRSHead vs) =
  mkTag "drs" ++ [ "variables" .= map formatVar vs ]
exprHeadToPairs (SDRSHead rs) =
  mkTag "sdrs" ++ [ "relations" .= rs ]
exprHeadToPairs MergeHead =
  mkTag "merge"
exprHeadToPairs (AlfaHead s) =
  mkTag "alfa" ++ [ "name" .= s ]
  
condHeadToPairs :: CondHead -> [Pair]
condHeadToPairs (SimpleHead n vs) =
  mkTag "simple" ++ [ "name" .= n, "variables" .= map formatVar vs ]
condHeadToPairs (EqHead v1 v2) =
  mkTag "equals" ++ [ "variables" .= [formatVar v1, formatVar v2] ]
condHeadToPairs (PropHead v) =
  mkTag "proposition" ++ [ "variable" .= formatVar v ]
condHeadToPairs (CardHead v cp) =
  mkTag "cardinality" ++ [ "variable" .= formatVar v, "proposition" .= cp ]
condHeadToPairs NotHead =
  mkTag "not"
condHeadToPairs PossibHead =
  mkTag "possible"
condHeadToPairs NecessHead =
  mkTag "necessary"
condHeadToPairs OrHead =
  mkTag "or"
condHeadToPairs ImplHead =
  mkTag "implies"
condHeadToPairs WhQHead =
  mkTag "wh-question"
condHeadToPairs (TimexHead v s) =
  mkTag "time" ++ [ "variable" .= v, "time" .= s ]
  
constHeadToPairs :: ConstituentHead -> [Pair]
constHeadToPairs (ConstituentHead v) =
  mkTag "constituent" ++ [ "variable" .= formatVar v ]


instance FromJSON DRT where
  parseJSON (Object v) = do
    tag <- v .: "tag" :: Parser Text
    h <- parseHead tag v
    label <- parseLabel v
    subs <- parseSubs v
    return $ DRT label h subs
  parseJSON _ = mzero

parseHead :: Text -> Object -> Parser DRTHead
parseHead tag v =
  case tag of
    "variable" ->
      fmap Expr $ VExpHead <$> (parseVar <$> v .: "variable")
    "lambda" -> 
      fmap Expr $ AbsHead <$> (parseVar <$> v .: "variable")
    "application" ->
      pure (Expr AppHead)
    "drs" ->
      fmap Expr $ DRSHead <$> (map parseVar <$> v .: "variables")
    "sdrs" ->
      fmap Expr $ SDRSHead <$> v .: "relations"
    "merge" ->
      pure (Expr MergeHead)
    "alfa" ->
      fmap Expr $ AlfaHead <$> v .: "name"
    "simple" ->
      fmap Cond $ SimpleHead <$> v .: "name" <*> (map parseVar <$> v .: "variables")
    "equals" -> do
      vs <- map parseVar <$> v .: "variables"
      if (length vs == 2)
        then return (Cond $ EqHead (vs !! 0) (vs !! 1))
        else throw $ FatalException "Invalid variable count for \"equals\" condition"
    "proposition" ->
      fmap Cond $ PropHead <$> (parseVar <$> v .: "variable")
    "cardinality" ->
      fmap Cond $ CardHead <$> (parseVar <$> v .: "variable") <*> v .: "proposition"
    "not" ->
      pure (Cond NotHead)
    "possible" ->
      pure (Cond PossibHead)
    "necessary" ->
      pure (Cond NecessHead)
    "or" -> 
      pure (Cond OrHead)
    "implies" ->
      pure (Cond ImplHead)
    "wh-question" ->
      pure (Cond WhQHead)
    "time" ->
      fmap Cond $ TimexHead <$> (parseVar <$> v .: "variable") <*> v .: "time"
    "constituent" ->
      fmap ELCS.DRT.Const $ ConstituentHead <$> (parseVar <$> v .: "variable")
    _ -> throw $ FatalException ("Invalid tag for DRT expression: " ++ unpack tag)
  
parseLabel :: Object -> Parser MRLabel
parseLabel v = do
  a <- v .: "alignment"
  t <- v .: "type"
  return $ MRLabel { mrLabelAlignment = a, mrLabelType = t }

{-
parseSubs :: Object -> Parser [DRT]
parseSubs v = (++) <$> subs <*> ((++) <$> conds <*> consts)
  where subs     = v .:? "sub-expressions" .!= []
        conds    = v .:? "conditions"      .!= []
        consts   = v .:? "constituents"    .!= []
-}
  
parseSubs :: Object -> Parser [DRT]
parseSubs v = v .: "sub-expressions" -- todo make this fail if any of the keys is present but has an error
          <|> v .: "conditions"
          <|> v .: "constituents"
          <|> pure []


-- Relation

instance ToJSON Relation where
  toJSON (Relation s vs) =
    object $ mkTag "relation" ++ [ "name" .= s, "variables" .= map formatVar vs ]
    
instance FromJSON Relation where
  parseJSON (Object v) =
    (v .: "tag" ==! "relation") *>
      (Relation <$> v .: "name" <*> (map parseVar <$> v .: "variables"))
  parseJSON _ = mzero
  
  
-- CardProp

instance ToJSON CardProp where
  toJSON (CardEq i) =
    object $ mkTag "equals" ++ [ "value" .= show i ]
    
instance FromJSON CardProp where
  parseJSON (Object v) =
    (v .: "tag" ==! "equals") *>
      (CardEq <$> (read <$> v .: "value"))
  parseJSON _ = mzero
  
  
-- Type

instance ToJSON Type where
  toJSON (Arrow t1 t2) =
    object $ mkTag "arrow" ++ [ "left" .= t1, "right" .= t2 ]
  toJSON (Var i) =
    object $ mkTag "variable" ++ [ "name" .= i ]
  toJSON Entity = object $ mkTag "entity"
  toJSON Truth = object $ mkTag "truth"
  
instance FromJSON Type where
  parseJSON (Object v) = do
    tag <- v .: "tag" :: Parser Text
    case tag of
      "arrow" -> Arrow <$> v .: "left" <*> v .: "right"
      "variable" -> Var <$> v .: "name"
      "entity" -> pure Entity
      "truth" -> pure Truth
      _ -> throw $ FatalException ("Invalid tag for expression type: " ++ unpack tag)
  parseJSON _ = mzero
  

-- ItemStats

$(deriveJSON defaultOptions{fieldLabelModifier = (drop 4)} ''ItemStats)
  
