{-# LANGUAGE TemplateHaskell, DeriveGeneric #-}
{-|
  Implements syntactic splitting.

  A /syntactic split/ divides a 'CCGItem' into two items that can be recombined by applying a
  combinator. In order to do this, two decisions must be made: Identifying a splitting index within
  the tokens of the item (easy) and performing a /semantic split/, i.e. splitting the semantic
  component of the item into two parts accordingly (hard).
  
  Note that the generation of syntactic categories is implemented here so as to not require any
  decisions. The assignment of syntactic categories to generated items is determined by the type
  of split being performed as well as the semantic type of the split meaning representations.
  
  The functions in this module implement the easy part and take a 'SemSplitter' to perform the
  semantic split. Since the semantic splitter in general produces a list of semantic splits, this
  leads to just as many splits being extracted for a given split index.
-}
module ELCS.CCG.Split (
  Split(..)
, splitSubItems
  
  -- * Splitting CCG items
, splitItem
, splitItemAt
  
  -- * Inference of syntactic categories
, inferSplitCats
) where

import ELCS.MR.Split
import ELCS.MR.Type
import ELCS.DRT
import ELCS.CCG.CCG
import ELCS.CCG.Semantics
import ELCS.Util

import Data.Maybe
import Control.Exception
import Control.DeepSeq.Generics
import GHC.Generics(Generic)


-- | A pair of CCG items that have been split by a combinator
data Split m = Split
  { splitCombinator :: Combinator
  , splitLeft :: CCGItem m
  , splitRight :: CCGItem m
  } deriving (Show, Eq, Generic)
  
instance (NFData m) => NFData (Split m) where rnf = genericRnf
  
-- | Puts both of a split’s items in a list.
splitSubItems :: Split m -> [CCGItem m]
splitSubItems s = [splitLeft s, splitRight s]


-- Splitting CCG items

-- | Performs all possible splits of a given item.
--
--   Searches across all split indices and applies @splitter@ to every index.
splitItem :: SemSplitter DRT -> CCGItem DRT -> [Split DRT]
splitItem splitter item =
  if length (ccgTokens item) < 2 then [] else let
    toks = ccgTokens item
    splitPoints = [1 .. length toks - 1]
    splitLists = map (splitItemAt splitter item) splitPoints
    splits = concat splitLists 
  in splits
  
-- | Performs all possible splits of a given item, at a specific split point.
splitItemAt :: SemSplitter DRT -> CCGItem DRT -> Int -> [Split DRT]
splitItemAt splitter item sp =
  let
    toks = ccgTokens item
    sem = ccgSemantics item
    syn = ccgSyntax item
  in assert ((length toks >= 2) && (sp > 0) && (sp < length toks)) $
  let
    semSplits = splitter sp sem
    synCats = map (inferSplitCats syn) semSplits
    tokSplit = (take sp toks, drop sp toks)
    items = map (mkSplitCCGItems sp) (zip3 (repeat tokSplit) synCats semSplits)
  in items
  
-- | Does the organisational work of putting tokens, syntax and semantics together the right way
--   for a split.
mkSplitCCGItems :: Int
                -> (([String], [String]), (SynCat, SynCat), SemSplit DRT)
                -> Split DRT
mkSplitCCGItems sp ((ltoks,rtoks), (hcat,ccat), semSplit) =
  let
    -- Reorder the tokens to match the combinator that was applied
    dir = combinatorDirection (semSplitCombinator semSplit)
    (htoks,ctoks) = case dir of
                      Forward -> (ltoks,rtoks)
                      Backward -> (rtoks,ltoks)
    (hf,cf) = case dir of
                Forward -> (0,sp)
                Backward -> (sp,0)
    hsem = shiftAlignments (-hf) $
           forceInferTypes $ makeVariablesUnique $ normalForm $ semSplitHead semSplit
    csem = shiftAlignments (-cf) $
           forceInferTypes $ makeVariablesUnique $ normalForm $ semSplitComplement semSplit
    hitem = CCGItem { ccgTokens = htoks, ccgSyntax = hcat, ccgSemantics = hsem }
    citem = CCGItem { ccgTokens = ctoks, ccgSyntax = ccat, ccgSemantics = csem }
    split = case dir of
              Forward -> Split (semSplitCombinator semSplit) hitem citem
              Backward -> Split (semSplitCombinator semSplit) citem hitem
  in split
                       
                       
-- Inference of syntactic categories of split results
  
-- | Given a desired resulting category, infers the syntactic categories for a split of a
--   given combinator.
inferSplitCats :: SynCat           -- ^ The resulting type of the combined items.
               -> SemSplit DRT     -- ^ The split.
               -> (SynCat, SynCat) -- ^ The syntactic categories.
inferSplitCats rcat spl = let
    comb  = semSplitCombinator spl
    (h,c) = (makeVariablesUnique $ semSplitHead spl, makeVariablesUnique $ semSplitComplement spl)
    cInfered = case inferTypes c of
                 Right c' | isCompleteType (fromJust $ drtType c') -> c'
                          | otherwise -> throw $ FatalException $
                                           "inferSplitCats: Incomplete type " ++
                                           show (fromJust $ drtType c') ++
                                           " for expression " ++ show c
                 Left e -> throw $ FatalException ("inferSplitCats: Type error (" ++ e ++ ")")
    cats = case comb of
             FA -> inferAppCats Forward rcat (h,cInfered)
             BA -> inferAppCats Backward rcat (h,cInfered)
  in cats
  
inferAppCats :: CombinatorDirection -> SynCat -> (DRT, DRT) -> (SynCat, SynCat)
inferAppCats dir rcat (_h,c) = let
    ctype = fromJust $ drtType c
    ccat = semToSynType ctype 
    hcat = case dir of 
             Forward -> ForwardSlash rcat ccat
             Backward -> BackwardSlash rcat ccat
  in (hcat,ccat)
