{-# LANGUAGE FlexibleContexts, TemplateHaskell #-}
{-|
  Implements recursive syntactic splitting.

  Recursive splitting means that every split that is returned by the splitting procedure
  is in turn subjected to further splitting. The recursion terminates if an item cannot be
  split further. This generally has one of two reasons:

  1. The semantic splitter is unable to split the meaning representation given its settings.
  2. The item only contains a single token and thus cannot be split.

  For now, the semantic splitter is pre-configured to an alignment-restricted single-branching
  recursive splitter.
-}
module ELCS.CCG.SplitRecursive (
  Options(..)
, ELCS.CCG.SplitRecursive.defaultOptions
, ExtractionResult
, ItemSet
, extract
, resultToGrammar
, resultToSplits
, resultItemCount
) where

import ELCS.CCG
import ELCS.DRT
import qualified ELCS.DRT.SplitOpportunistic as SO

import qualified Data.Map as Map
--import Control.Exception(assert)
--import Debug.Trace
import Control.Monad
import Control.Monad.Reader
import Control.Monad.Error
import qualified Data.Aeson.TH as AesTH

-- Options

-- | Options that may be passed to the extraction algorithm.
--   Most of the options actually concern the splitting algorithm that is used as a sub-procedure.
data Options = Options
  { splitOpportunistic :: SO.Options -- ^ Options for the semantic splitter.
  , maxItemCount :: Int              -- ^ The maximum number of items to extract per sentence.
  }

-- | Default options.
defaultOptions :: Options
defaultOptions = Options
  { splitOpportunistic = SO.defaultOptions
  , maxItemCount = 10000
  }

$(AesTH.deriveJSON AesTH.defaultOptions ''Options)


-- Main interface

-- | Extraction results in a packed structure containing:
--
--   * all grammar items that were extracted
--
--   * the set of splits that were performed on each item
--
--   * a designated root item representing the original input sentence
--
--   The functions 'resultToGrammar' and 'resultToSplits' can be used to obtain the relevant
--   useful data structures.
type ExtractionResult = (CCGItem DRT, ItemSet)

-- | Performs grammar extraction.
extract :: Options                        -- ^ Options for the extraction process.
        -> Derivation                     -- ^ Derivation of the sentence.
        -> Either String ExtractionResult -- ^ The extraction result or an error.
extract opts d = do
  root <- rootItem d
  items <- runReader
             (runErrorT
               (extractCachedLimited (maxItemCount opts) (Map.empty) root))
             opts
  return (root, items)


-- Processing extraction results

-- | Extracts a grammar from an item set. The resulting grammar is sorted in an ascending order.
resultToGrammar :: ExtractionResult -> (Grammar DRT)
resultToGrammar = map fst . Map.toAscList . snd

-- | Extracts a list of splits.
resultToSplits :: ExtractionResult -> [(CCGItem DRT, Split DRT)]
resultToSplits (_, items) = concat $ map entryToSplits $ Map.toList items

entryToSplits :: (CCGItem DRT, [Split DRT]) -> [(CCGItem DRT, Split DRT)]
entryToSplits (a, ss) = zip (repeat a) ss

-- | Returns the number of items that are in the extracted grammar.
resultItemCount :: ExtractionResult -> Int
resultItemCount = Map.size . snd


-- Splitting with memoization

-- | We use a map to avoid performing duplicate splits and sort the grammar along the way.
--
--   The keys to the map are the extracted items. Under each key, every split that was
--   generated from the item is stored.
type ItemSet = Map.Map (CCGItem DRT) [Split DRT]

-- | Monad within which splitting functions run.
type Splitter = ErrorT String (Reader Options)

-- | Calls extractCachedGen but aborts if the grammar size exceeds the limit
extractCachedLimited :: Int -> ItemSet -> CCGItem DRT -> Splitter ItemSet
extractCachedLimited limit c item =
  if Map.size c <= limit
    then extractCachedGen (extractCachedLimited limit) c item
    else throwError ("Exceeded maximum item count of " ++ show limit)

-- | Inserts all of an item’s recursive splits into a set.
--   This algorithm uses 'c' to prune duplicate parts of the recursion tree.
--   Every item generated is inserted into 'c', and if it is inserted for the first time,
--   it is split recursively.
extractCachedGen :: (ItemSet -> CCGItem DRT -> Splitter ItemSet)
                    -- ^ The split procedure to call for recursion. Having this as a parameter
                    --   allows us to extend the basic procedure, e.g. by an item limit.
                 -> ItemSet          -- ^ The chart of items that were already generated.
                 -> CCGItem DRT      -- ^ The item to split.
                 -> Splitter ItemSet -- ^ The modified item set.
extractCachedGen rec c item =
  asks splitOpportunistic >>= \splitOpts ->
  case item `Map.member` c of
    True  -> return c
    False -> do
      let splits = splitItem (SO.splitAll splitOpts) item
      let items = concat $ map splitSubItems splits
      let c' = Map.insert item splits c
      --trace ("split " ++ show (Set.size c) ++ "(" ++ show (length children) ++ ") " ++
      --       show (ccgTokens item)) $
      foldM rec c' items
