{-# LANGUAGE FlexibleInstances, FlexibleContexts, MultiParamTypeClasses,
             ScopedTypeVariables, DeriveDataTypeable, DeriveGeneric, DefaultSignatures #-}
{-|
  Implements abstract syntax trees for meaning representations in discourse representation theory
  (DRT).

  Auxiliary tools such as constructor functions, a zipper implementation, and operations such
  as reductions and renamings are also provided within this module.

  Specifically, this implementation is for the extended flavour of DRT used in the Groningen
  Meaning Bank (GMB), with segmented discourse representation structures (SDRS), merge expressions,
  λ-calculus, and several other extensions.
-}
module ELCS.DRT.AST (
-- * DRT data structures
-- $dataStructures

  DRT(..)
, DRTHead(..)
, ExprHead(..)
, CondHead(..)
, ConstituentHead(..)
, Relation(..)
, CardProp(..)

-- * Zipper

, Zipper
, Breadcrumb(..)
, zipperExpr
, zipperSetExpr
, zipperCrumbs
, toZipper
, fromZipper
, zipUp
, zipDown
, zipToChild
, zipLeft
, zipRight
, zipTop
, zipperSubIndex
, zipperSubs
, zipperParents
, leftSiblings
, rightSiblings
, spliceZippers

-- * Bindings and visibility

, directBindings
, isDirectlyBoundBy
, effectiveBindings
, varsUsedBy
, shadowsBinding
, isUsedBy
, isRightPropagating
, reinventBindings

-- * Constructor functions
-- $constructorFunctions

, mkVExp
, mkVExp'
, mkAbs
, mkAbs'
, mkApp
, mkApp'
, mkDRS
, mkDRS'
, mkSDRS
, mkSDRS'
, mkMerge
, mkMerge'
, mkAlfa
, mkAlfa'
, mkSimple
, mkSimple'
, mkEq
, mkEq'
, mkProp
, mkProp'
, mkCard
, mkCard'
, mkNot
, mkNot'
, mkPossib
, mkPossib'
, mkNecess
, mkNecess'
, mkOr
, mkOr'
, mkImpl
, mkImpl'
, mkWhQ
, mkWhQ'
, mkTimex
, mkTimex'
, mkConstituent
, mkConstituent'

-- * Extended constructor functions
, mkAbsChecked
, mkAppRenaming

-- * Sanity checking

, SyntaxError
, SyntaxErrorType(..)
, isSane
, sanityCheck

-- * Substitution

, substitute

-- * Renaming variables

, alphaRename
, renameFreeVar

-- * Reduction strategies

, normalForm
, reduceOuterLeft

-- * Beta reduction

, reduceBeta
, betaNormalForm

-- * Reduction of merge expressions

, reduceMerge
, mergeNormalForm
, deAlfa
, reduceAlfa

-- * Variable bindings

, deepBindings
, checkVariablesUnique
, makeVariablesUnique

-- * Accessing components of expressions

, drtAlignment
, drtSetAlignment
, drtSetAlignmentRecursive
, drtType
, drtSetType
, drtMaybeSetType
, isDRS
, drsVars
, drsConds
, shiftAlignments

) where

import ELCS.MR
import ELCS.MR.Label
import ELCS.MR.Type
import ELCS.Util

import Control.Exception
import Control.DeepSeq
import Control.DeepSeq.Generics
import qualified Control.Monad.State as State
import Data.List
import Data.Maybe
import qualified Data.Set as Set
import GHC.Generics(Generic)
--import Debug.Trace(trace,traceShow)


-- $dataStructures
-- To keep algorithms simple, the implementation of DRT expressions abstracts out the expression’s
-- tree structure. Expressions are therefore simple rose trees. Each node of an expression
-- contains two types of annotation:
--
--   1. a 'MRLabel' which may specify alignment and type information,
--
--   2. a 'DRTHead' which contains describes the semantic function of the node.
--
-- On the first level, 'DRTHead' distinguishes between three classes of nodes: Expressions,
-- DRT conditions, and SDRS constituents. On the next level, the concrete node types
-- present in the GMB flavour of DRT are defined.
--
-- The advantage of this design is that expressions can be treated uniformly without resorting
-- to huge case expressions, as we would have to if we were to literally implement the AST as
-- a sum-of-products type.
--
-- A major disadvantage is that there exist some syntactic constraints which cannot be encoded
-- into the data type.
--
--   * Many node types enforce a constraint on the number of sub-expressions. E.g., an application
--     node must have exactly two sub-nodes.
--
--   * There are constraints on the node classes: conditions must be children of a DRS, constituents
--     of an SDRS.
--
-- These constraints can be validated at runtime using 'sanityCheck'.

-- | The main data structure for DRT expressions.
--
--   Implements a tree where each node is labelled with an 'MRLabel' and a 'DRTHead'.
data DRT = DRT { drtLabel :: MRLabel,
                 drtHead :: DRTHead,
                 drtSubs :: [DRT] }
           deriving (Show, Eq, Ord, Read, Generic)

-- | Specifies the semantic function of a node in a DRT expression.
data DRTHead = Expr ExprHead
             | Cond CondHead
             | Const ConstituentHead
             deriving (Show, Eq, Ord, Read, Generic)

-- | Represents nodes corresponding to expressions.
data ExprHead = VExpHead  Var
              | AbsHead   Var
              | AppHead
              | DRSHead   [Var]
              | SDRSHead  [Relation]
              | MergeHead
              | AlfaHead  String
              deriving (Show, Eq, Ord, Read, Generic)

-- | Represents nodes corresponding to DRS conditions.
data CondHead = SimpleHead String [Var]
              | EqHead     Var    Var
              | PropHead   Var
              | CardHead   Var    CardProp
              | NotHead
              | PossibHead
              | NecessHead
              | OrHead
              | ImplHead
              | WhQHead
              | TimexHead  Var    String
              deriving (Show, Eq, Ord, Read, Generic)

-- | Represents nodes corresponding to SDRS constituents.
data ConstituentHead = ConstituentHead Var
                       deriving (Show, Eq, Ord, Read, Generic)

-- | An SDRS relation
data Relation = Relation String [Var]
                deriving (Show, Eq, Ord, Read, Generic)

-- | A cardinality proposition about a referent
data CardProp = CardEq Integer           -- Cardinality equals
                deriving (Show, Eq, Ord, Read, Generic)


instance NFData DRT where rnf = genericRnf
instance NFData DRTHead where rnf = genericRnf
instance NFData ExprHead where rnf = genericRnf
instance NFData CondHead where rnf = genericRnf
instance NFData ConstituentHead where rnf = genericRnf
instance NFData Relation where rnf = genericRnf
instance NFData CardProp where rnf = genericRnf


-- Zipper implementation

-- | A breadcrumb which can be used to reconstruct a given node in an unzipped expression.
--   crumbLeft and crumbRight are both stacks of sub-trees, with the top elements being the direct
--   siblings of the sub-tree under focus. The sub-tree under focus itself is not represented in
--   the breadcrumb, but in the Zipper.
data Breadcrumb = Breadcrumb { crumbLabel :: MRLabel,
                               crumbHead  :: DRTHead,
                               crumbLeft  :: [DRT],
                               crumbRight :: [DRT] }
                  deriving (Eq, Show)

-- | A zipper for DRT expressions.
type Zipper = (DRT, [Breadcrumb])

-- | Reconstructs an expression given a breadcrumb and a sub-tree.
unCrumb :: Breadcrumb -> DRT -> DRT
unCrumb b e = DRT (crumbLabel b) (crumbHead b) ((reverse $ crumbLeft b) ++ [e] ++ (crumbRight b))

-- | Retrieves the expression under focus.
zipperExpr :: Zipper -> DRT
zipperExpr = fst

-- | Replaces the expression under focus.
zipperSetExpr :: Zipper -> DRT -> Zipper
zipperSetExpr (_,bs) e = (e,bs)

zipperCrumbs :: Zipper -> [Breadcrumb]
zipperCrumbs = snd

-- | Returns a zipper that corresponds to the root of the expression.
toZipper :: DRT -> Zipper
toZipper e = (e, [])

-- | Reconstructs the expression represented by the zipper.
fromZipper :: Zipper -> DRT
fromZipper z = zipperExpr $ zipTop z

-- | Moves the zipper up one level. Returns Nothing if it is already at the top.
zipUp :: Zipper -> Maybe Zipper
zipUp (_, [])   = Nothing
zipUp (e, b:bs) = Just (unCrumb b e, bs)

-- | Moves the zipper down one level. Returns Nothing if the current sub-tree has no children.
--   The leftmost sub-tree will be under focus.
zipDown :: Zipper -> Maybe Zipper
zipDown ((DRT l h (e:es)), bs) = let
    b = Breadcrumb { crumbLabel = l, crumbHead = h, crumbLeft = [], crumbRight = es }
  in Just (e, b:bs)
zipDown ((DRT _ _ _), _) = Nothing

-- | Zip to the @i@th child of the expression under focus.
zipToChild :: Int -> Zipper -> Maybe Zipper
zipToChild i z =
  let subs = zipperSubs z
  in if length subs > i
       then Just $ subs !! i
       else Nothing

-- | Moves the zipper to the left sibling.
zipLeft :: Zipper -> Maybe Zipper
zipLeft (e, b:bs) =
  case (crumbLeft b, crumbRight b) of
    ([],_)       -> Nothing
    ((l:ls),rs)  -> Just $ (l, (b { crumbLeft = ls, crumbRight = e:rs }):bs)
zipLeft _ = Nothing

-- | Moves the zipper to the left sibling.
zipRight :: Zipper -> Maybe Zipper
zipRight (e, b:bs) =
  case (crumbLeft b, crumbRight b) of
    (_,[])       -> Nothing
    (ls,(r:rs))  -> Just $ (r, (b { crumbLeft = e:ls, crumbRight = rs }):bs)
zipRight _ = Nothing

-- | Moves the zipper to the top of the expression.
zipTop :: Zipper -> Zipper
zipTop z = case zipUp z of
             Just z' -> zipTop z'
             Nothing -> z

-- | Returns the child index of the expression currently under focus.
zipperSubIndex :: Zipper -> Int
zipperSubIndex (_,bs) =
  case bs of
    (b:_) -> length (crumbLeft b)
    [] -> 0


zipperSubs :: Zipper -> [Zipper]
zipperSubs z = case zipDown z of
                 Just z' -> [z'] ++ rightSiblings z'
                 Nothing -> []

zipperParents :: Zipper -> [Zipper]
zipperParents z = case zipUp z of
                    Just z' -> [z'] ++ zipperParents z
                    Nothing -> []

leftSiblings :: Zipper -> [Zipper]
leftSiblings z = case zipLeft z of
                   Just z' -> leftSiblings z' ++ [z']
                   Nothing -> []

rightSiblings :: Zipper -> [Zipper]
rightSiblings z = case zipRight z of
                    Just z' -> [z'] ++ rightSiblings z'
                    Nothing -> []


-- | Inserts the expression represented by the inner zipper into the focus of the outer zipper.
--   The resulting zipper still points to the same sub tree as the inner zipper, but the surrounding
--   expression has been expanded.
spliceZippers :: Zipper -- ^ The outer zipper.
              -> Zipper -- ^ The inner zipper.
              -> Zipper -- ^ The resulting zipper.
spliceZippers (_,obs) (e,ibs) = (e, ibs ++ obs)


-- Bindings and visibility

-- | Returns the set of variables that are bound directly by an expression’s head.
directBindings :: DRT -> [Var]
directBindings e =
  case drtHead e of
    Expr (AbsHead v)   -> [v]
    Expr (SDRSHead rs) -> foldr (\(Relation _ vs) vs' -> union vs vs') [] rs
    Expr (DRSHead vs)  -> vs
    _ -> []

isDirectlyBoundBy :: Var -> DRT -> Bool
isDirectlyBoundBy v e = v `elem` directBindings e

-- | Returns the set of bindings that are available in sub-expressions of e, including
--   right-propagated bindings.
effectiveBindings :: DRT -> [Var]
effectiveBindings e | isRightPropagating (drtHead e) =
  assert (length (drtSubs e) == 2) $ -- All right-propagating expressions are binary
  directBindings e `union` effectiveBindings (drtSubs e !! 0)
effectiveBindings e | otherwise = directBindings e

varsUsedBy :: DRT -> [Var]
varsUsedBy e =
  case drtHead e of
    Expr (VExpHead v)         -> [v]
    Expr (SDRSHead _)         -> [] -- No usages because all used variables are also bound
    Cond (SimpleHead _ vs)    -> vs
    Cond (EqHead v1 v2)       -> [v1, v2]
    Cond (PropHead v)         -> [v]
    Cond (CardHead v _)       -> [v]
    Cond (TimexHead v _)      -> [v]
    Const (ConstituentHead v) -> [v]
    _                         -> []
    
-- | Indicates whether a node shadows a binding of a given name.
shadowsBinding :: Var -> DRT -> Bool
shadowsBinding v e = v `elem` (effectiveBindings e)
  
isUsedBy :: Var -> DRT -> Bool
isUsedBy v e = v `elem` (varsUsedBy e)


-- | Some DRT constructions have a right-propagating property: They make the variables bound
--   in their first sub-expression available in the second sub-expression (these constructions
--   are always binary). This function is an indicator for this property.
isRightPropagating :: DRTHead -> Bool
isRightPropagating e =
  case e of
    Expr MergeHead -> True
    Expr (AlfaHead _) -> True
    Expr _ -> False
    Cond ImplHead -> True
    Cond WhQHead -> True
    Cond _ -> False
    Const (ConstituentHead _) -> False


-- MR instance

instance MR DRT where
  subExprs = drtSubs

  mrLabel = drtLabel

  vars e = effectiveBindings e `union` varsUsedBy e `union` (foldr union [] $ map vars $ drtSubs e)

  freeVars e =
    let
      freeInSub =
        case isRightPropagating (drtHead e) of
          True -> let [l,r] = drtSubs e
                  in union (freeVars l) (freeVars r \\ effectiveBindings l)
          False -> foldr union [] $ map freeVars $ drtSubs e
    in varsUsedBy e `union` (freeInSub \\ directBindings e)


-- Constructing DRT expressions

-- $constructorFunctions
-- In order to make constructing DRT expressions a little easier on the eyes, convenience
-- constructor functions are provided for all types of expressions. These generally come in two
-- versions:
--
--   1. The @mk<HeadType>@ funktions take a 'MRLabel', then all arguments for the head, and then the
--      children.
--   2. The @mk<HeadType>'@ functions which omit the label and always insert an 'emptyMRLabel'
--      instead.
--
-- In addition, a few extended constructor functions are also available which perform some
-- additional work.

mkVExp :: MRLabel -> Var -> DRT
mkVExp l v = DRT l (Expr (VExpHead v)) []

mkVExp' :: Var -> DRT
mkVExp' = mkVExp emptyMRLabel

mkAbs :: MRLabel -> Var -> DRT -> DRT
mkAbs l v e = DRT l (Expr (AbsHead v)) [e]

mkAbs' :: Var -> DRT -> DRT
mkAbs' = mkAbs emptyMRLabel

mkApp :: MRLabel -> DRT -> DRT -> DRT
mkApp l e1 e2 = DRT l (Expr AppHead) [e1,e2]

mkApp' :: DRT -> DRT -> DRT
mkApp' = mkApp emptyMRLabel

mkDRS :: MRLabel -> [Var] -> [DRT] -> DRT
mkDRS l vs cs = DRT l (Expr (DRSHead vs)) cs

mkDRS' :: [Var] -> [DRT] -> DRT
mkDRS' = mkDRS emptyMRLabel

mkSDRS :: MRLabel -> [DRT] -> [Relation] -> DRT
mkSDRS l cs rs = DRT l (Expr (SDRSHead rs)) cs

mkSDRS' :: [DRT] -> [Relation] -> DRT
mkSDRS' = mkSDRS emptyMRLabel

mkMerge :: MRLabel -> DRT -> DRT -> DRT
mkMerge l e1 e2 = DRT l (Expr (MergeHead)) [e1,e2]

mkMerge' :: DRT -> DRT -> DRT
mkMerge' = mkMerge emptyMRLabel

mkAlfa :: MRLabel -> String -> DRT -> DRT -> DRT
mkAlfa l n e1 e2 = DRT l (Expr (AlfaHead n)) [e1,e2]

mkAlfa' :: String -> DRT -> DRT -> DRT
mkAlfa' = mkAlfa emptyMRLabel

-- Conditions
mkSimple :: MRLabel -> String -> [Var] -> DRT
mkSimple l n vs = DRT l (Cond (SimpleHead n vs)) []

mkSimple' :: String -> [Var] -> DRT
mkSimple' = mkSimple emptyMRLabel

mkEq :: MRLabel -> Var -> Var -> DRT
mkEq l v1 v2 = DRT l (Cond (EqHead v1 v2)) []

mkEq' :: Var -> Var -> DRT
mkEq' = mkEq emptyMRLabel

mkProp :: MRLabel -> Var -> DRT -> DRT
mkProp l v e = DRT l (Cond (PropHead v)) [e]

mkProp' :: Var -> DRT -> DRT
mkProp' = mkProp emptyMRLabel

mkCard :: MRLabel -> Var -> CardProp -> DRT
mkCard l v p = DRT l (Cond (CardHead v p)) []

mkCard' :: Var -> CardProp -> DRT
mkCard' = mkCard emptyMRLabel

mkNot :: MRLabel -> DRT -> DRT
mkNot l e = DRT l (Cond NotHead) [e]

mkNot' :: DRT -> DRT
mkNot' = mkNot emptyMRLabel

mkPossib :: MRLabel -> DRT -> DRT
mkPossib l e = DRT l (Cond PossibHead) [e]

mkPossib' :: DRT -> DRT
mkPossib' = mkPossib emptyMRLabel

mkNecess :: MRLabel -> DRT -> DRT
mkNecess l e = DRT l (Cond NecessHead) [e]

mkNecess' :: DRT -> DRT
mkNecess' = mkNecess emptyMRLabel

mkOr :: MRLabel -> DRT -> DRT -> DRT
mkOr l e1 e2 = DRT l (Cond OrHead) [e1,e2]

mkOr' :: DRT -> DRT -> DRT
mkOr' = mkOr emptyMRLabel

mkImpl :: MRLabel -> DRT -> DRT -> DRT
mkImpl l e1 e2 = DRT l (Cond (ImplHead)) [e1,e2]

mkImpl' :: DRT -> DRT -> DRT
mkImpl' = mkImpl emptyMRLabel

mkWhQ :: MRLabel -> DRT -> DRT -> DRT
mkWhQ l e1 e2 = DRT l (Cond (WhQHead)) [e1,e2]

mkWhQ' :: DRT -> DRT -> DRT
mkWhQ' = mkWhQ emptyMRLabel

mkTimex :: MRLabel -> Var -> String -> DRT
mkTimex l v s = DRT l (Cond (TimexHead v s)) []

mkTimex' :: Var -> String -> DRT
mkTimex' = mkTimex emptyMRLabel

-- Constituent
mkConstituent :: MRLabel -> Var -> DRT -> DRT
mkConstituent l v e  = DRT l (Const (ConstituentHead v)) [e]

mkConstituent' :: Var -> DRT -> DRT
mkConstituent' = mkConstituent emptyMRLabel


-- Extended constructor functions

-- | Creating a non-vacuous abstraction: an assertion checks whether the variable abstracted over
--   actually occurs.
mkAbsChecked :: MRLabel -> Var -> DRT -> DRT
mkAbsChecked l v e = assertMsg ("\nchecked abstraction: " ++ show v ++
                                " not free in (" ++ show e ++ ")")
                               (v `elem` freeVars e) $
                     mkAbs l v e

-- | Combines two expressions to form a new application expression. The variables in @e2@ are
--   renamed to be disjunct of those in @e1@, so that 'checkVariablesUnique' holds for the result.
mkAppRenaming :: MRLabel -> DRT -> DRT -> DRT
mkAppRenaming l e1 e2 = let
    e2' = (reinventBindings (vars e1) e2)
    app = mkApp l e1 e2'
  in assert (checkVariablesUnique app) app


-- Sanity checking

-- | Describes an error that is found by 'sanityCheck'.
--
--   The zipper points to the exact location of the error (i.e., the node that violated a
--   constraint), while the second component of the pair describes the error that was found.
type SyntaxError = (Zipper, SyntaxErrorType)

-- | Defines the types of syntax errors that can be found. See 'sanityCheck' for definitions of the
--   error types.
data SyntaxErrorType =
  NestingError -- ^ The nesting condition was violated.
  deriving (Show,Eq)

-- | Returns True iff 'sanityCheck' finds no errors.
isSane :: DRT -> Bool
isSane e = sanityCheck e == Nothing

-- | Checks for the following constraints on a DRT expression:
--
--     1. “Nesting condition”: The node type matches the node type that is expected due to its
--        parent. Children of DRS nodes must be conditions, children of SDRS nodes must be
--        constituents. All other nodes must be expressions.
--
--   The constraint that every node has the expected number of children should be checked here,
--   but this is currently not implemented.
sanityCheck :: DRT -> Maybe SyntaxError
sanityCheck = sanityCheckZ . toZipper

sanityCheckZ :: Zipper -> Maybe SyntaxError
sanityCheckZ z = let
    nesting = checkNesting z
    errors = catMaybes [nesting]
  in if null errors then Nothing else Just $ head errors

checkNesting :: Zipper -> Maybe SyntaxError
checkNesting z = let
    e = zipperExpr z
    pass = case drtHead e of
             Expr (DRSHead _) -> all isCond (drtSubs e)
             Expr (SDRSHead _) -> all isConst (drtSubs e)
             Expr _ -> all isExpr (drtSubs e)
             _ -> all isExpr (drtSubs e)
    localError = if pass then Nothing else Just (z, NestingError)
    subErrors = case zipDown z of
                  Just z' -> checkNesting z'
                  Nothing -> Nothing
    rightErrors = case zipRight z of
                    Just z' -> checkNesting z'
                    Nothing -> Nothing
    errors = catMaybes $ [localError] ++ [subErrors] ++ [rightErrors]
  in if null errors then Nothing else Just $ head errors

isCond :: DRT -> Bool
isCond e = case drtHead e of
             Cond _ -> True
             _ -> False

isConst :: DRT -> Bool
isConst e = case drtHead e of
              Const _ -> True
              _ -> False

isExpr :: DRT -> Bool
isExpr e = case drtHead e of
             Expr _ -> True
             _ -> False


-- Substitution

-- | Performs a substitution of a free variable by an expression.
--   Read: “Substitute @v@ by @r@ in @e@”.
--
--   Variables bound in @r@ are renamed so as not to clash with those in @e@.
--
--   This method can perform two different kinds of substitutions.
--
--     * Variable-variable substitutions: If @r@ is a 'VExp', every occurrence of @v@
--       is substituted by the variable referenced in @r@.
--
--     * Variable-expression substitutions: If @r@ is anything other than a 'VExp',
--       every occurrence of @v@ is substituted by the expression @r@. In order for this
--       substitution to be valid, every occurrence of @v@ must be in a 'VExp'.
substitute :: Var -> DRT -> DRT -> DRT
substitute v r e =
  doSubstitute v r (reinventBindings (vars r) e) -- Rename to avoid capturing

-- | Helper for 'substitute'. Performs substitution on an already renamed expression.
doSubstitute :: Var -> DRT -> DRT -> DRT
doSubstitute v r e =
  if shadowsBinding v e
    then e
    else
      case drtHead e of
        Expr (VExpHead v') | v == v'   -> r
                           | otherwise -> e
        _ -> let
               e' = substituteVarUsage v r e
               subs' = map (doSubstitute v r) (drtSubs e')
             in e' { drtSubs = subs' }

-- | Helper for 'substitute'.
--   Unconditionally substitutes variable usages in the head by an expression.
--   Note that the replacement expression must be a VExp unless all occurrences of the placeholder
--   variable are VExps. (We cannot substitute complex expressions in predicate arguments, e.g.)
substituteVarUsage :: Var -> DRT -> DRT -> DRT
substituteVarUsage v r e =
    case drtHead r of
      Expr (VExpHead v') -> substituteVarUsageByVar v v' e
      _                  -> assert checkVarUnused e
  where
    checkVarUnused = if (v `isUsedBy` e)
                       then throw $ FatalException
                              ("Cannot replace " ++ (show v) ++ " by " ++ (show r) ++
                               " in " ++ (show e))
                       else True

-- | Helper for 'substitute'.
--   Substitutes variable usages in the head by a variable.
substituteVarUsageByVar :: Var -> Var -> DRT -> DRT
substituteVarUsageByVar v v' (DRT l h es) =
  let
    h' = case h of
      Expr (VExpHead v1)           -> Expr (VExpHead (replace v v' v1))
      --Expr (SDRSHead rs)           -> Expr (SDRSHead (map (relSubstituteVar v v') rs))
      Cond (SimpleHead n vs)       -> Cond (SimpleHead n (map (replace v v') vs))
      Cond (EqHead v1 v2)          -> Cond (EqHead (replace v v' v1) (replace v v' v2))
      Cond (PropHead v1)           -> Cond (PropHead (replace v v' v1))
      Cond (CardHead v1 n)         -> Cond (CardHead (replace v v' v1) n)
      Cond (TimexHead v1 n)        -> Cond (TimexHead (replace v v' v1) n)
      Const (ConstituentHead v1)   -> Const (ConstituentHead (replace v v' v1))
      _                            -> h
  in DRT l h' es

-- | Replaces a variable in a relation.
relSubstituteVar :: Var -> Var -> Relation -> Relation
relSubstituteVar v v' (Relation n vs) = Relation n (map (replace v v') vs)


-- Renaming bound variables

-- | Renames bound variables in @e@ so they do not contain any variables from @rvars@.
reinventBindings :: [Var] -> DRT -> DRT
reinventBindings rvars e =
  let
    result = zipperExpr $
             State.evalState (reinventBindingsHelper rvars (toZipper e))
                             (initMRBuilder (union rvars (vars e)))
  in assertMsg ("reinventBindings: correctness check" ++
                "\nrvars: " ++ show rvars ++ "\ne: " ++ show e ++
                "\nresult: " ++ show result)
               (null (rvars `intersect` deepBindings result))
               result

reinventBindingsHelper :: [Var] -> Zipper -> MRBuilder Zipper
reinventBindingsHelper rvars z =
  --trace ("reinventBindingsHelper " ++ show rvars ++ " " ++ show (zipperExpr z)) $
  assert (not $ containsDuplicatesOrd $ directBindings $ zipperExpr z) $
  do
    let bound = directBindings (zipperExpr z)
    let inventIfNeeded v@(n,_) = if v `elem` rvars then mrInventVar n else return v
    newVars <- mapM inventIfNeeded bound
    let mapping = filter (uncurry (/=)) $ zip bound newVars
    let renamed = foldr (\(o,n) prz -> alphaRename o n prz) z mapping
    let leftChild = zipDown renamed
    childrenRenamed <- case leftChild of
                         Just c -> do reinv <- reinventBindingsInChildren rvars c
                                      return (fromJust $ zipUp reinv)
                         Nothing -> return renamed
    return childrenRenamed

reinventBindingsInChildren :: [Var] -> Zipper -> MRBuilder Zipper
reinventBindingsInChildren rvars z =
  do
    renamed <- reinventBindingsHelper rvars z
    siblingsRenamed <- case zipRight renamed of
                         Just z' -> reinventBindingsInChildren rvars z'
                         Nothing -> return renamed
    return siblingsRenamed

-- | Does a bound replacement of variable v by r in e. r must be bound by e. Variables must be
--   unique in the whole expression.
alphaRename :: Var -> Var -> Zipper -> Zipper
alphaRename v r z =
  --trace ("alphaRename " ++ show v ++ " " ++ show r ++ " " ++ show z) $
  assertMsg ("alphaRename: Variable to be renamed is not bound!\ne: " ++ show (zipperExpr z))
            (v `isDirectlyBoundBy` (zipperExpr z)) $
  let
    e = zipperExpr z
    head' = case drtHead e of
              Expr (AbsHead bv) -> assert (v == bv) (Expr (AbsHead r))
              Expr (DRSHead vs) -> Expr (DRSHead (map (replace v r) vs))
              Expr (SDRSHead rs) -> Expr (SDRSHead (map (relSubstituteVar v r) rs))
              _ -> throw $ FatalException "alphaRename: Missing binding expression type in case"
    z' = zipperSetExpr z (e { drtHead = head' })
    kidsRenamed = renameFreeVarInSubs v r z' 0
    outerRenamed = propagateRenameFreeVar v r kidsRenamed
  in outerRenamed

propagateRenameFreeVar :: Var -> Var -> Zipper -> Zipper
propagateRenameFreeVar v r z =
  case zipUp z of
    Just p ->
      if isRightPropagating (drtHead (zipperExpr p))
        then let idx = zipperSubIndex z
                 siblingsRenamed = renameFreeVarInSubs v r p (idx + 1)
             in fromJust $ zipToChild idx $ propagateRenameFreeVar v r siblingsRenamed
        else z
    Nothing -> z

-- | Renames a free variable in all sub expressions starting at index i.
renameFreeVarInSubs :: Var -> Var -> Zipper -> Int -> Zipper
renameFreeVarInSubs v r z i = let
    e = zipperExpr z
    subs = drtSubs e
    toBeRenamed = drop i subs
    renamed = map (renameFreeVar v r) toBeRenamed
    subs' = take i subs ++ renamed
  in zipperSetExpr z (e { drtSubs = subs' })

-- | Renames a free variable in an expression.
renameFreeVar :: Var -> Var -> DRT -> DRT
renameFreeVar v r e
    | shadowsBinding v e = e
    | otherwise       = e { drtHead = head', drtSubs = subs' }
      where
        subs = drtSubs e
        head' = drtHead $ substituteVarUsageByVar v r e
        subs' = map (renameFreeVar v r) subs


-- Reduction strategies

normalForm :: DRT -> DRT
normalForm = mergeNormalForm . deAlfa . betaNormalForm

-- | Perform a single outermost-leftmost reduction, return Nothing if the expression is in
--   normal form.
reduceOuterLeft :: (DRT -> Maybe (DRT)) -> DRT -> Maybe (DRT)
reduceOuterLeft red e = do
    reduced <- reduceOuterLeftZ red (toZipper e)
    return (fromZipper reduced)

-- | Perform a single outermost-leftmost reduction in the focused sub-tree.
--   Return Nothing if the expression is in normal form.
reduceOuterLeftZ :: (DRT -> Maybe (DRT)) -> Zipper -> Maybe (Zipper)
reduceOuterLeftZ red zipper@(e,bs) =
  case red e of                 -- Is this already a redex?
    Just e' -> Just (e',bs)     -- If yes, just reduce and stop recursing
    Nothing -> zipDown zipper   -- If no, try the siblings from left to right
               >>= trySiblings
               >>= zipUp
  where
    trySiblings :: Zipper -> Maybe (Zipper)
    trySiblings z = case reduceOuterLeftZ red z of  -- Was there a redex inside this sibling?
                      Just z' -> Just z'            -- If yes, stop recursing
                      Nothing -> zipRight z         -- If no, try the next sibling
                                 >>= trySiblings


-- Beta reduction

-- | Reduce an expression if it is a beta redex, return Nothing otherwise.
--
--   A beta redex is of the form @(λx.e)f@, i.e. an application with an abstraction as its left
--   child. The reduction is then performed according to λ-calculus, i.e. every occurrence of @x@
--   in @e@ is substituted by @f@ using 'substitute'.
reduceBeta :: DRT -> Maybe (DRT)
reduceBeta (DRT _ _ [DRT _ (Expr (AbsHead v)) [e], r]) = Just $ substitute v r e
reduceBeta _ = Nothing


-- | Performs outermost-leftmost reductions until e is in normal form.
betaNormalForm :: DRT -> DRT
betaNormalForm e = case reduceOuterLeft reduceBeta e of
                     Just e' -> betaNormalForm e'
                     Nothing -> e


-- Merge reduction

-- | Reduces a merge expression.
--
--   A merge can be reduced in one of two cases:
--
--     * Both of its children are DRSs. Then the two DRSs are merged into one, meaning that
--       the union of their universes and conditions is built.
--
--     * The left child is a DRS and the right child is a merge, whose left child is another DRS.
--       Then, the two DRSs can also be merged. The scheme is as follows: @(a;(b;c)) => (ab;c)@.
reduceMerge :: DRT -> Maybe DRT
reduceMerge e = case drtHead e of
                  Expr (MergeHead) -> combine (drtSubs e)
                  _                 -> Nothing
  where
    combine :: [DRT] -> Maybe DRT
    combine [e1,e2] =
      case drtHead e2 of
        -- Directly merge two DRS
        Expr (DRSHead _) -> drsMerge e1 e2
        -- Merge (e1;e2l) in an expression of the form (e1;(e2l;e2r))
        Expr (MergeHead) -> let
            e2l = (drtSubs e2) !! 0
            e2r = (drtSubs e2) !! 1
          in do
               merged <- (drsMerge e1 e2l)
               return $ mkMerge (drtLabel e) merged e2r
        _ -> Nothing
    combine _ = throw $ FatalException "reduceMerge: wrong sub expression count of merge"

-- | Merges two DRSs. The variables of @e2@ are renamed to not conflict with those of @e1@.
--
--   If either @e1@ or @e2@ is not a DRS, @Nothing@ is returned.
drsMerge :: DRT -> DRT -> Maybe DRT
drsMerge e1 e2 = let
    e2' = reinventBindings (vars e1) e2
  in if isDRS e1 && isDRS e2 then Just (forceDrsMerge e1 e2') else Nothing

-- | Helper for @drsMerge@. Merges two DRSs.
forceDrsMerge :: DRT -> DRT -> DRT
forceDrsMerge e1 e2 =
  assert (isDRS e1 && isDRS e2) $
  let
    e1Bindings = fromJust $ drsVars e1
    e2Bindings = fromJust $ drsVars e2
    e1Conds = fromJust $ drsConds e1
    e2Conds = fromJust $ drsConds e2
  in assertMsg ("These DRSs are expected to be binding-disjunct:\ne1: " ++ show e1 ++
                "\ne2: " ++ show e2)
               (null (e1Bindings `intersect` e2Bindings)) $ let
    vs'    = e1Bindings ++ e2Bindings
    conds' = e1Conds ++ e2Conds
  in mkDRS (drtLabel e1) vs' conds'

-- | Reduces all merge expressions.
mergeNormalForm :: DRT -> DRT
mergeNormalForm e = case reduceOuterLeft reduceMerge e of
                      Just e' -> mergeNormalForm e'
                      Nothing -> e


-- Converting Alfa nodes to Merge nodes

-- | The normal form for 'reduceAlfa': Replaces all 'AlfaHead' nodes by 'MergeHead' nodes.
deAlfa :: DRT -> DRT
deAlfa e = case reduceOuterLeft reduceAlfa e of
             Just e' -> deAlfa e'
             Nothing -> e

-- | Replaces an 'AlfaHead' node by a 'MergeHead' node.
--
--   'AlfaHead' is semantically equivalent to 'MergeHead', so we keep things simple and only deal
--   with one kind of merge.
reduceAlfa :: DRT -> Maybe (DRT)
reduceAlfa e = case drtHead e of
                 Expr (AlfaHead _) -> Just $ e { drtHead = Expr (MergeHead) }
                 _                 -> Nothing


-- General utility functions

-- | If 'x == y', returns 'r', otherwise 'y'. This is a handy shorthand for replacement of
--   all occurrences of a given component (usually a sub-expression) by another.
replace :: Eq a => a -> a -> a -> a
replace x r y | x == y    = r
              | otherwise = y


-- Uniqueness of variable bindings

-- | Returns all variables bound somewhere in @e@.
deepBindings :: DRT -> [Var]
deepBindings e = directBindings e ++ (concat $ map deepBindings $ drtSubs e)

-- | Checks whether all variables are unique. If there is more than one binding using the same
--   variable name, variables are not unique.
checkVariablesUnique :: DRT -> Bool
checkVariablesUnique = not . containsDuplicatesOrd . deepBindings

-- | Returns an alpha-equivalent expression that contains no two distinct variables with the
--   same name.
makeVariablesUnique :: DRT -> DRT
makeVariablesUnique e =
  -- trace ("makeVariablesUnique (" ++ show e ++ ")") $
  let
    z = State.evalState (doMakeVariablesUnique $ toZipper e) emptySeenVarsState
    e' = zipperExpr z
  in -- Correctness postcondition
     assertMsg ("makeVariablesUnique: Incorrect result\n" ++
                "input: " ++ show e ++ "\noutput: " ++ show e')
               (checkVariablesUnique e') $
     -- If variables were already unique, the expression should stay the same
     assertMsg ("makeVariablesUnique: Changed expression although variables already unique\n" ++
                "before: " ++ show e ++ "\nafter: " ++ show e')
               (not (checkVariablesUnique e) || e == e') $
     e'

-- | Implements 'makeVariablesUnique'.
--
--   The implementation is done using a simple State monad to keep track of variables seen
--   so far. This function contains the logic of walking through the expression
--   in a depth-first, left-to-right manner.
doMakeVariablesUnique :: Zipper -> SeenVars Zipper
doMakeVariablesUnique z = do
  -- Step 1: Rename locally bound variables
  let e = zipperExpr z
  seen <- State.get
  let renamed = reinventBindings (Set.toList seen) e
  let z1 = zipperSetExpr z renamed
  
  State.put $ Set.union seen (Set.fromList $ directBindings $ zipperExpr z1)

  -- Step 2: recurse into children
  z2 <- case zipDown z1 of
          Just sub -> do sub' <- doMakeVariablesUnique sub
                         return $ fromJust $ zipUp sub'
          Nothing -> return z1
  
  -- Step 3: go to right sibling
  z3 <- case zipRight z2 of
          Just sibling -> doMakeVariablesUnique sibling
          Nothing -> return z2
  return z3

type SeenVarsState = Set.Set Var

emptySeenVarsState :: SeenVarsState
emptySeenVarsState = Set.empty

type SeenVars = State.State SeenVarsState


-- Accessing components of expressions

-- | Returns the set of words a given node is aligned to.
drtAlignment :: DRT -> Maybe [Int]
drtAlignment = mrLabelAlignment . drtLabel

-- | Sets the alignment of a node.
drtSetAlignment :: [Int] -> DRT -> DRT
drtSetAlignment al e = let
    label = drtLabel e
    label' = label { mrLabelAlignment = Just al }
  in e { drtLabel = label' }

-- | Sets the alignment of a node and all sub-nodes.
drtSetAlignmentRecursive :: [Int] -> DRT -> DRT
drtSetAlignmentRecursive al e = let
    subs' = map (drtSetAlignmentRecursive al) $ drtSubs e
    e' = e { drtSubs = subs' }
  in drtSetAlignment al e'

-- | Returns the semantic type of a node.
drtType :: DRT -> Maybe Type
drtType = mrLabelType . drtLabel

-- | Sets the semantic type of a node.
drtSetType :: Type -> DRT -> DRT
drtSetType t e = let
    label = drtLabel e
    label' = label { mrLabelType = Just t }
  in e { drtLabel = label' }

-- | Sets the type of a node if @mt@ is @Just t@. Otherwise, leaves the node alone.
drtMaybeSetType :: Maybe Type -> DRT -> DRT
drtMaybeSetType mt e = case mt of
                         Just t  -> drtSetType t e
                         Nothing -> e

-- | Tests whether a given node is a DRS node.
isDRS :: DRT -> Bool
isDRS e = case drtHead e of
            Expr (DRSHead _) -> True
            _ -> False

-- | If @e@ is a DRS node, returns its universe.
drsVars :: DRT -> Maybe [Var]
drsVars e =
  case drtHead e of
    Expr (DRSHead vs) -> Just vs
    _                 -> Nothing

-- | If @e@ is a DRS node, returns its conditions.
drsConds :: DRT -> Maybe [DRT]
drsConds e =
  case drtHead e of
    Expr (DRSHead _) -> Just (drtSubs e)
    _                -> Nothing


-- | Shifts alignments in an expression by a constant offset.
shiftAlignments :: Int -> DRT -> DRT
shiftAlignments o e =
  let
    reAligned = case drtAlignment e of
                  Just al -> drtSetAlignment (map (+o) al) e
                  Nothing -> e
    subs' = map (shiftAlignments o) (drtSubs e)
  in reAligned { drtSubs = subs' }
