{-# LANGUAGE TemplateHaskell #-}

{-|
  This module implements opportunistic splitting of DRT expressions.
  
  The term “opportunistic” refers to the fact that we do not force extraction of a split for any
  given split point. Instead, a given split may be rejected if it would yield an item that is
  considered too complex, making it implausible.
  
  The basic interface function 'split' accepts a combinator and a split definition as its arguments.
  These parameters partition the sentence's word indices into two sets:
  /head words/ and /complement words/. E.g., for the 'FA' combinator, words to the left of the
  split point are considered to be head words and words starting from the split point are
  complement words. This is flipped for the 'BA' combinator. In general, this mapping is
  determined by the directionality of the combinator (see 'ELCS.CCG.combinatorDirection').
  
  This in turn partitions the nodes of an expressions into three sets. Nodes that are aligned
  exclusively to head words become /head nodes/, while nodes that are aligned exclusively to
  complement words are called /complement nodes/. Nodes without an alignment, as well as nodes that
  are aligned to both head and complement words, become /unaligned nodes/ (they effectively have
  their alignments deleted).
  
  To understand the splitting strategy, first consider its non-recursive subset. In this variant,
  a (possible) split is given by an expression along with a designated /split node/ that is part of
  the expression. A non-recursive split is /valid/ if the split node governs all complement nodes
  but no head nodes.
  
  We then extend this set by allowing recursive splitting. A recursive split is given by a list
  of split nodes, which are required to lie on a common path from the root of the expression.
  Another way of stating this requirement is that the tree governed by any split node is an
  (indirect) subtree of the next higher split node’s tree, thus forming a nested hierarchy
  of subtrees. For a recursive split to be /valid/, we require the following:
  
    * All complement nodes are governed by the highest split node.
    
    * Construct a new recursive split by taking the highest split node as root and keeping all but
      the highest split node. Swap the roles of head and complement nodes.
      If there is only one split node left, require that the corresponding
      non-recursive split be valid. Otherwise, require that the recursive split be valid.
      
  In effect, this partitions our expression into a set of nested sub-trees alternatingly “coloured”
  as containing either no head nodes or no complement nodes.
  
  Note that this recursion is unary-branching: This is the fundamental criterion of complexity we
  apply.
  
  This criterion makes the bookkeeping required to transform a split into an actual pair of
  λ-expressions (a /head expression/ and a /complement expression/) relatively simple.
  In the non-recursive case, the following are required:
  
    * Replace the sub-tree governed by the split node (“the split tree”) by a new variable, and
      abstract over it in the generated head expression.
      
    * In the complement expression, abstract over any variables that are free in the split tree.
      In the head expression, apply these variables to the newly generated variable to pass them
      along correctly.
      
  Applying this principle to recursive splits is slightly more tricky, but conceptually not any
  more complicated.
-}
module ELCS.DRT.SplitOpportunistic (
  Options(..)
, defaultOptions
, splitAll
, split
) where

import ELCS.CCG
import ELCS.DRT
import ELCS.MR
import ELCS.MR.Label
import ELCS.MR.Type
import ELCS.MR.Split
import ELCS.Util

import Data.Maybe
import Data.List
import Control.Exception(assert)
import Control.Monad.Reader
import qualified Data.Aeson.TH as AesTH
--import Debug.Trace

type DRTZipper = Zipper


-- Configuration

data Options = Options
  { maxArrowCount :: Int
  , maxGeneratedFreeVars :: Int
  , maxSplitCount :: Int
  , maxRecursionDepth :: Int
  , maxUnalignedConditions :: Int
  }
  
defaultOptions :: Options
defaultOptions = Options
  { maxArrowCount = 8
  , maxGeneratedFreeVars = 3
  , maxSplitCount = 1000
  , maxRecursionDepth = 4
  , maxUnalignedConditions = 10
  }
  
$(AesTH.deriveJSON AesTH.defaultOptions ''Options)
  
type Splitter = Reader Options


-- | Split with every available combinator
splitAll :: Options -> Int -> DRT -> [SemSplit DRT]
splitAll opts idx e = split opts FA idx e ++ split opts BA idx e

-- | Performs opportunistic splitting of an expression given a split index.
split :: Options -> Combinator -> Int -> DRT -> [SemSplit DRT]
split opts comb idx e = runReader (doSplit comb idx e) opts

-- A note on terminology: To abstract from “left” and “right” split halves (which have
-- meaning in connection with the tokens list), we instead talk about “outer” and “inner”, as well
-- as “head” and “complement” alignments. The head is the first argument of the CCG combinator, the
-- complement the second. During the splitting process, these roles may become reversed, which is
-- why many functions take a pair of 'BinaryAlignment's. The first of these is understood to be
-- the outer alignment, the second the inner alignment.

doSplit :: Combinator                          -- ^ The combinator to reverse in this split. 
        -> Int                                 -- ^ Word index to split at.
        -> DRT                                 -- ^ Meaning representation to split.
        -> Splitter [SemSplit DRT]
doSplit comb idx e =
  -- trace ("--------> split " ++ show comb ++ " " ++ show idx ++ " (" ++ show e ++ ")") $
  do
    hcPairs <- case comb of
                 FA -> splitApp 0 (Head,Complement) (mkSplitType ForwardSplit idx) e
                 BA -> splitApp 0 (Head,Complement) (mkSplitType BackwardSplit idx) e
    let toSemSplit (h,c) = SemSplit { semSplitCombinator = comb,
                                      semSplitHead = forceInferTypes $ makeVariablesUnique h,
                                      semSplitComplement = forceInferTypes $ makeVariablesUnique c }
    let splits = map toSemSplit hcPairs
    limitSplitCount =<< filterM isAcceptedSplit splits
  
-- | Returns an empty list if the list of splits is too long.
limitSplitCount :: [a] -> Splitter [a]
limitSplitCount spls =
  ask >>= \opts ->
  return $
    -- lazily determine if list length is greater than allowed
    if null (drop (maxSplitCount opts) spls)
      then spls
      else []
  
-- | Used to filter out some splits
isAcceptedSplit :: SemSplit DRT -> Splitter Bool
isAcceptedSplit s =
  ask >>= \opts ->
  return $
    exprsSatisfy ((<= maxArrowCount opts) . arrowCount . fromJust . drtType)
  where exprsSatisfy p = (p $ semSplitHead s) && (p $ semSplitComplement s)

-- Constructor function for SplitType
mkSplitType :: SplitDirection -> Int -> SplitType
mkSplitType dir sp =
  SplitType { splitDirection = dir,
              splitPoint = sp }

-- | Prepares a DRS for splitting.
--
-- DRSs are nodes with potentially many children, all of which may or may not contain alignments
-- to the “inner” side of the split. DRS splitting groups these conditions together and generates
-- a corresponding merge expression, which allows splitting all of them out in a single operation.
--
-- Since there may be unaligned conditions, a number of splittings are valid. Every possible
-- grouping with regard to these conditions is generated.
--
-- Precondition: We're dealing with a DRS.
-- todo: we might consider splitting in both directions, i.e. putting the inner conditions
-- either on the left or the right side of the merge. Currently, they are always on the right.
drsSplits :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper
          -> Splitter [DRTZipper]
drsSplits al s drsz =
  assert (isDRS (zipperExpr drsz)) $
  ask >>= \opts ->
  let
    drs = zipperExpr drsz
    vs = fromJust $ drsVars drs
    conds = drtSubs drs
  -- Reject splitting a singleton DRS to avoid endless loop
  in if length (conds) <= 1 then return [] else
  let
    governsOuter = filter (governsOuterNodes al s) conds
    governsInner = filter (governsInnerNodes al s) conds
    governsBoth = governsOuter `intersect` governsInner
    outerConds = governsOuter \\ governsBoth
    innerConds = governsInner
    unalignedConds = ((conds \\ outerConds) \\ innerConds)
  -- Abort if we have too many unaligned conditions
  in if length unalignedConds > (maxUnalignedConditions opts) then return [] else
  let
    unalignedSubsets = subsequences unalignedConds
    splitOuter = map (outerConds ++) unalignedSubsets
    splitInner = map (innerConds ++) $
                 map (unalignedConds \\) unalignedSubsets
    csplits = filter (\(o,i) -> not (null o || null i)) $
              zip splitOuter splitInner
    merges = map (uncurry $ drsSplitMerge vs) csplits
    mergeZippers = map (zipperSetExpr drsz) merges
    mergeNodes = map (fromJust . zipRight . fromJust . zipDown) mergeZippers
  in return mergeNodes
  
-- | Generates a merge expression and two DRS expressions given two sets of conditions.
drsSplitMerge :: [Var] -> [DRT] -> [DRT] -> DRT
drsSplitMerge vs oconds iconds = let
    ovs = vs `intersect` (foldr union [] $ map vars oconds)
    ivs = vs \\ ovs
    odrs = mkDRS unalignedMRLabel ovs oconds
    idrs = mkDRS unalignedMRLabel ivs iconds
  in mkMerge unalignedMRLabel odrs idrs


-- Implementation of application splits
  
-- Strategy:
-- Find the lowest node that governs all inner nodes.
-- If this is the root of the expression, return (λx.x,e) <- e.g., semantics for '.'.
-- If no inner nodes exist, reject splitting.
-- Recurse. Go up and repeat as long as no new outer nodes are being governed.
-- For each recursion result (f,a), return (λx.abstractFreeVars(e[f/xa]),applyFreeVars(f))

-- | Produces a split for the application combinator.
splitApp :: Int -> (BinaryAlignment,BinaryAlignment) -> SplitType -> DRT -> Splitter [(DRT, DRT)]
splitApp recDepth al s e =
  assertMsg ("splitApp: is not supposed to have free variables: " ++ show e) (null $ freeVars e) $
  ask >>= \opts ->
  if recDepth <= (maxRecursionDepth opts)
    then do
      result <- doSplitApp recDepth al s (toZipper e)
      return $ (assert (all (isValidAppSplit e) result))
               result
    else return []

isValidAppSplit :: DRT -> (DRT,DRT) -> Bool
isValidAppSplit _e (h,c) =
  {-alphaEquivalent (normalForm $ mkAppT emptyMRLabel h c) (normalForm e) &&-}
  null (freeVars h) && null (freeVars c)

doSplitApp :: Int -> (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper -> Splitter [(DRT, DRT)]
doSplitApp recDepth al s z =
  assert (null $ zipperCrumbs z) $
  assert (checkVariablesUnique (zipperExpr z)) $
  if not (governsInnerNodesZ al s z) -- Reject splitting if there are no inner nodes at all
    then return []
    else
      do
        let splitNodes = findSplitNodes al s z
        splits <- mapM (splitAtNode recDepth al s) splitNodes
        return (concat splits)
    
-- | Abstracts the variables that are free at a given node.
--
-- The resulting expression will be alpha-equivalent to the old expression, but the sub-expression
-- focused by the zipper will not have any free variables.
--
-- This operation is performed as a preparation for splitting out a sub-expression.
abstractFreeVars :: DRTZipper -> DRTZipper
abstractFreeVars z = let
    e = zipperExpr z
    fv = freeVars e
    zAbs = toZipper $ forceInferTypes $ reinventBindings (vars (zipperExpr $ zipTop z)) $
           mkAbsChain e fv
    zApp = mkOuterAppChain zAbs (map (mkVExp unalignedMRLabel) $ reverse fv)
    z' = spliceZippers z zApp
  in --assertMsg ("\nBefore: " ++ show (zipperExpr $ zipTop z) ++
     --           "\nAfter: " ++ show (zipperExpr $ zipTop z'))
     --          (alphaEquivalent (normalForm $ zipperExpr $ zipTop z')
     --                           (normalForm $ zipperExpr $ zipTop z))
     assertMsg ("prepareSplitInner: Variables not unique in result" ++
                "\nz: " ++ show z ++ "\nz': " ++ show z')
               (checkVariablesUnique (zipperExpr $ zipTop z')) $
     z'

-- | Once a split node has been identified, this function performs the split.
--
-- If the node is a condition, the corresponding DRS is split as required. If the node
-- is an SDRS constituent, no splitting is performed.
splitAtNode :: Int -> (BinaryAlignment,BinaryAlignment) -> SplitType -> (DRTZipper,Int)
            -> Splitter [(DRT,DRT)]
splitAtNode recDepth al s (z,d) =
  ask >>= \opts ->
     -- todo The first part of the condition is overly restrictive: We want to stop splitting
     -- if the root is the lowest split node both for the current and the reversed split direction
     -- (because this means that we will never find a final split node).
  if (d == 0 && governsInnerNodesZ al s z && governsOuterNodesZ al s z) ||
     ((length $ freeVars $ zipperExpr z) > (maxGeneratedFreeVars opts))
    then return []
    else
      case drtHead $ zipperExpr z of
        Expr _  -> splitAtExpr recDepth al s z
        Cond _  -> do -- drsSplits expects a zipper to the DRS, so zip up
                     merges <- drsSplits al s (fromJust $ zipUp z)
                     splits <- mapM (splitAtExpr recDepth al s) merges
                     return (concat splits)
        Const _ -> return [] -- Reject splitting out whole constituents

-- | Splits an expression without any pre-processing.
--
-- Decides whether a simple or a recursive split is required.
splitAtExpr :: Int -> (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper
            -> Splitter [(DRT, DRT)]
splitAtExpr recDepth al@(o,i) s z
  | governsOuterNodesZ al s z = do
      -- If we cannot do a simple app split, we attempt a unary recursive app split.
      -- We reverse inner and outer roles of nodes and recurse.
      let z' = abstractFreeVars z
      subSplits <- splitApp (recDepth + 1) (i,o) s (zipperExpr z')
      let splits = catMaybes $ map (unaryAppSplit z') subSplits
      return splits
  | otherwise = return $ maybeToList $ finalSplitAtExpr z
  
-- | Does the grunt work of splitting an expression under application in the recursive case.
-- Two expressions that resulted from a split of the sub-expression under focus are also
-- taken as an argument.
unaryAppSplit :: DRTZipper            -- ^ The node at which to split
              -> (DRT,DRT)            -- ^ The sub-head and sub-complement to substitute into the split
              -> Maybe (DRT, DRT)     -- ^ Resulting head and complement
unaryAppSplit z (oSub, iSub) =
  assert (checkVariablesUnique iSub) $
  assert (checkVariablesUnique oSub) $
  assert (checkVariablesUnique (zipperExpr $ zipTop z)) $
  let
    plug = oSub
    iSubVar = inventVar "v" (vars oSub `union` vars iSub `union` vars (zipperExpr $ zipTop z))
    iSubRenamed = reinventBindings (vars (zipperExpr $ zipTop z)) iSub
    socketType = drtType $ zipperExpr z
    socket = drtMaybeSetType socketType $
             mkAppT unalignedMRLabel (mkVExp unalignedMRLabel iSubVar) iSubRenamed
    z' = zipperSetExpr z socket
    hull = mkAbsChecked unalignedMRLabel iSubVar $ zipperExpr $ zipTop z'
  in -- Can happen if we split out a part of an expression that has right-extending visibility:
     if (not (null $ freeVars hull)) then Nothing else
             assertMsg ("\nsplitAtExprRecursive: plug has free variables")
               (null $ freeVars plug) $
     assertMsg ("\nsplitAtExprRecursive failed\ne: (" ++ show (zipperExpr $ zipTop z) ++ ")\nhull: (" ++ show hull ++ ")\nplug: (" ++ show plug ++ ")")
               (isValidAppSplit (zipperExpr $ zipTop z) (hull,plug)) $
     assertMsg ("splitAtExprRecursive: vars not unique in z'\n" ++
                "z: " ++ show z ++ "\nz': " ++ show z')
               (checkVariablesUnique (zipperExpr $ zipTop z')) $
     assert (checkVariablesUnique hull) $
     assert (checkVariablesUnique plug) $
     Just (hull,plug)

-- | Performs the split for a given expression in the case that the inner part of the split
--   does not need to be split further.
finalSplitAtExpr :: DRTZipper            -- ^ The node at which to split.
                 -> Maybe (DRT, DRT)     -- ^ Resulting head and complement.
finalSplitAtExpr z =
  assert (checkVariablesUnique (zipperExpr $ zipTop z)) $
  let
    v = inventVar "v" (vars $ zipperExpr $ zipTop $ z)
    i = zipperExpr z
    fv = freeVars i
    plug = mkAbsChain i fv
    socketType = drtType $ zipperExpr z
    socket = drtMaybeSetType socketType $
             mkAppChain (mkVExp unalignedMRLabel v) (map (mkVExp unalignedMRLabel) $ reverse fv)
    z' = zipperSetExpr z socket
    hull = mkAbsChecked unalignedMRLabel v $ zipperExpr $ zipTop z'
  in -- Can happen if we split out a part of an expression that has right-extending visibility:
     if (not (null $ freeVars hull)) then Nothing else 
     assertMsg ("\nfinalSplitAtExpr failed\ne: (" ++ show (zipperExpr $ zipTop z) ++ ")\nhull: (" ++ show hull ++ ")\nplug: (" ++ show plug ++ ")")
               (isValidAppSplit (zipperExpr $ zipTop z) (hull,plug)) $
     assert (checkVariablesUnique hull) $
     assert (checkVariablesUnique plug) $
     assert (checkVariablesUnique (zipperExpr $ zipTop z')) $
     Just (hull,plug)

findSplitNodes :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper -> [(DRTZipper,Int)]
findSplitNodes al s z = let
    lowest = findLowestSplitNode al s z
    nodes = [lowest] ++ splitNodeParents al s lowest
    nodes' = filter (isSubstantialSplit al s) nodes
  in {-trace ("SplitNodes found: " ++ show (length nodes) ++ " – " ++ show nodes)-}
     nodes'

-- | Filters out any splits that only split off abstraction nodes at the head of the expression.
--   This fixes an infinite recursion problem in connection with prepareSplitInner.
isSubstantialSplit :: (BinaryAlignment,BinaryAlignment) -> SplitType -> (DRTZipper,Int) -> Bool
isSubstantialSplit _al _s ((_e,[]),_) = True -- Exception: Allow identity splits
isSubstantialSplit _al _s ((_e,bs),_d) = not $ all (isAbsHead . crumbHead) bs
  where isAbsHead h = case h of
                        Expr (AbsHead _) -> True
                        _                -> False
  
-- | Finds the lowest node that governs all inner nodes.
findLowestSplitNode :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper -> (DRTZipper,Int)
findLowestSplitNode = doFindLowestSplitNode 0

doFindLowestSplitNode :: Int -> (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper -> (DRTZipper,Int)
doFindLowestSplitNode d al s z = let
    -- If exactly one node matches this filter, we can descend to it, otherwise abort.
    candidates = filter (governsInnerNodesZ al s) $ zipperSubs z
  in case candidates of
       [c] -> doFindLowestSplitNode (d+1) al s c
       _   -> (z,d)
    
splitNodeParents :: (BinaryAlignment,BinaryAlignment) -> SplitType -> (DRTZipper,Int) -> [(DRTZipper,Int)]
splitNodeParents al s (z,d) = let
    -- Get all parents that are split nodes. This is a bit tricky because we need to know
    -- the relevant child from which we are ascending. So we filter over the children and then
    -- zip up.
    parents = map (fromJust . zipUp) $ takeWhile (parentIsSplitNode al s) ([z] ++ zipperParents z)
    -- Finally, we need to annotate the correct depth values.
  in zip parents (takeWhile (>=0) [d-1,d-2..])

parentIsSplitNode :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper -> Bool
parentIsSplitNode al s z = let
    pz = zipUp z
    left = leftSiblings z
    right = rightSiblings z
  in case pz of
       Just parent -> (not $ isOuterNodeZ al s parent) &&
                      all (not . (governsOuterNodesZ al s)) left &&
                      all (not . (governsOuterNodesZ al s)) right
       _           -> False


governsOuterNodes :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRT -> Bool
governsOuterNodes al s e = isOuterNode al s e ||
                           (any (governsOuterNodes al s) $ drtSubs e)
                           
governsInnerNodes :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRT -> Bool
governsInnerNodes al s e = isInnerNode al s e ||
                           (any (governsInnerNodes al s) $ drtSubs e)
                         
isOuterNode :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRT -> Bool
isOuterNode (outer,_) s e =
  case drtAlignment e of
    Just alignment -> Just outer == (evalAlignment s alignment)
    Nothing -> False

isInnerNode :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRT -> Bool
isInnerNode (_,inner) s e =
  case drtAlignment e of
    Just alignment -> Just inner == (evalAlignment s alignment)
    Nothing -> False

governsOuterNodesZ :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper -> Bool
governsOuterNodesZ al s z = governsOuterNodes al s (zipperExpr z)

governsInnerNodesZ :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper -> Bool
governsInnerNodesZ al s z = governsInnerNodes al s (zipperExpr z)
                         
isOuterNodeZ :: (BinaryAlignment,BinaryAlignment) -> SplitType -> DRTZipper -> Bool
isOuterNodeZ al s z = isOuterNode al s (zipperExpr z)

-- | Creates a chain of application nodes with e at the head
mkAppChain :: DRT -> [DRT] -> DRT
mkAppChain e = foldl' (mkAppT unalignedMRLabel) e

-- | Creates a chain of application nodes on the outside of the focused expression.
mkOuterAppChain :: Zipper -> [DRT] -> Zipper
mkOuterAppChain z = foldl' mkOuterApp z

mkOuterApp :: Zipper -> DRT -> Zipper
mkOuterApp (e,bs) a = let
    e' = (reinventBindings (vars a) e)
    crumb = Breadcrumb { crumbHead = Expr AppHead
                       , crumbLabel = unalignedMRLabel
                       , crumbLeft = []
                       , crumbRight = [a] }
  in (e',bs ++ [crumb])

mkAbsChain :: DRT -> [Var] -> DRT
mkAbsChain e = foldl' (flip (mkAbsChecked unalignedMRLabel)) e