{-# LANGUAGE Arrows, NoMonomorphismRestriction, FlexibleInstances #-}

-- This module outputs DRT expressions as HTML documents for debugging.
module ELCS.DRT.ToHTML where

import ELCS.DRT.AST
import ELCS.MR
import ELCS.Util

import Data.List
import Text.XML.HXT.Core
import Text.XML.HXT.Arrow.XmlState.RunIOStateArrow
import Control.Exception

outputDRT :: String -> DRT -> IO ()
outputDRT f s = runWithInput (writeExpToFile f) s
  
runWithInput :: IOSLA (XIOState ()) a XmlTree -> a -> IO ()
runWithInput arrow e = do
  _ <- runIOSLA arrow (initialState ()) e
  return ()
  
writeExpToFile :: String -> IOSLA (XIOState s) DRT XmlTree
writeExpToFile f = fromLA toHtmlA >>> writeHtmlFile f
                  
writeHtmlFile :: String -> IOSLA (XIOState s) XmlTree XmlTree
writeHtmlFile f = traceMsg 1 ("Writing to " ++ f) >>>
                  writeDocument [withIndent yes, withOutputXHTML] f
  
toHtml :: DRT -> XmlTree
toHtml e = head $ runLA toHtmlA e

toHtmlA :: LA DRT XmlTree
toHtmlA =
  root [] [
    selem "html" [
      selem "head" [
        aelem "link" [ attr "rel" (txt "stylesheet")
                     , attr "href" (txt "gmbstyle/style.css")
                     , attr "type" (txt "text/css")]
      ]
    , selem "body" [
        mkelem "div" [ attr "class" (txt "xdrs") ] [
          writeExp
        ]
      ]
    ]
  ]

writeExp :: LA DRT XmlTree
writeExp = proc e -> do
  exprHead <- case drtHead e of
    Expr h -> this -< h
    _      -> arr (\e -> throw $ FatalException ("writeExp: argument is not an expression:\n" ++ (show e))) -< e

  content <- case exprHead of
    (VExpHead  _)  -> writeVar    -< e
    (AbsHead   _)  -> writeAbs    -< e
    AppHead        -> writeApp    -< e
    MergeHead      -> writeMerge  -< e
    (DRSHead    _) -> writeDRS    -< e
    (SDRSHead   _) -> writeSDRS   -< e
    (AlfaHead   _) -> writeAlfa   -< e
    
  returnA -< content
  
    
writeVar :: LA DRT XmlTree
writeVar =
  arr drtHead >>>
  proc (Expr (VExpHead v)) ->
    returnA <<< mkText -< formatVar v

writeAbs :: LA DRT XmlTree
writeAbs = text <+> body
  where
    text = arr drtHead >>>
           arr (\(Expr (AbsHead v)) -> ("λ" ++ (formatVar v) ++ ".")) >>>
           mkText
    body = arr drtSubs >>> arr head >>> writeExp

writeApp :: LA DRT XmlTree
writeApp = txt "(" <+> sub1 <+> txt "@" <+> sub2 <+> txt ")"
  where
    sub1 = arr drtSubs >>> arr (!!0) >>> writeExp
    sub2 = arr drtSubs >>> arr (!!1) >>> writeExp

writeMerge :: LA DRT XmlTree
writeMerge = txt "(" <+> sub1 <+> txt " ; " <+> sub2 <+> txt ")"
  where
    sub1 = arr drtSubs >>> arr (!!0) >>> writeExp
    sub2 = arr drtSubs >>> arr (!!1) >>> writeExp

writeDRS :: LA DRT XmlTree
writeDRS =
    mkelem "table" [ attr "class" (txt "drs") ] [
      selem "tr" [
        mkelem "td" [attr "class" (txt "indexlist")] [
          writeVars
        ]
      ]
    , selem "tr" [
        mkelem "td" [ attr "class" (txt "content") ] [
          selem "ul" [ writeConds ]
        ]
      ]
    ]
  where
    writeVars  = arr drtHead >>>
                 arr (\(Expr (DRSHead vs)) -> concat $ intersperse " " $ map formatVar vs) >>>
                 mkText
    writeConds = arr drtSubs >>> unlistA >>> writeCond
    
writeSDRS :: LA DRT XmlTree
writeSDRS =
    mkelem "table" [ attr "class" (txt "sdrs") ] [
      selem "tr" [
        mkelem "td" [ attr "class" (txt "constituents") ] [
          writeConsts ]
      ]
    , selem "tr" [
        mkelem "td" [ attr "class" (txt "relations") ] [
          selem "ul" [ writeRels ]
        ]
      ]
    ]
  where
    writeConsts = arr drtSubs >>> unlistA >>> writeConstituent
    writeRels = arr drtHead >>>
                arr (\(Expr (SDRSHead rs)) -> rs) >>>
                unlistA >>>
                writeRelation

writeAlfa :: LA DRT XmlTree
writeAlfa = txt "(" <+> sub1 <+> txt " *" <+> name <+> txt "* " <+> sub2 <+> txt ")"
  where
    name = arr drtHead >>> arr (\(Expr (AlfaHead n)) -> n) >>> mkText
    sub1 = arr drtSubs >>> arr (!!0) >>> writeExp
    sub2 = arr drtSubs >>> arr (!!1) >>> writeExp
                

writeCond :: LA DRT XmlTree
writeCond = proc e -> do
    condHead <- case drtHead e of
      Cond c -> this -< c
      _      -> throw $ FatalException "writeCond: argument is not a condition" -< ()
      
    content <- case condHead of
      (SimpleHead _ _) -> writeSimple -< e
      (EqHead _ _)     -> writeEq     -< e
      (PropHead _)     -> writeProp   -< e
      (CardHead _ _)   -> writeCard   -< e
      NotHead          -> writeNot    -< e
      PossibHead       -> writePossib -< e
      NecessHead       -> writeNecess -< e
      OrHead           -> writeOr     -< e
      ImplHead         -> writeImpl   -< e
      WhQHead          -> writeWhQ    -< e
      TimexHead _ _    -> writeTimex  -< e
      
    returnA <<< selem "li" [ this ] -< content

writeSimple :: LA DRT XmlTree
writeSimple = arr drtHead >>>
              arr (\(Cond (SimpleHead n vs)) -> (n ++ "(" ++ (argList vs) ++ ")")) >>>
              mkText
  where
    argList :: [Var] -> String
    argList = concat . intersperse "," . map formatVar
    
writeEq :: LA DRT XmlTree
writeEq = arr drtHead >>>
          arr (\(Cond (EqHead v1 v2)) -> (formatVar v1) ++ " = " ++ (formatVar v2)) >>>
          mkText 
    
writeProp :: LA DRT XmlTree
writeProp = text <+> sub
  where
    text = arr drtHead >>>
           arr (\(Cond (PropHead v)) -> (formatVar v) ++ ":") >>>
           mkText 
    sub = arr drtSubs >>> arr (!!0) >>> writeExp
      
writeCard :: LA DRT XmlTree
writeCard = proc e -> do
    c <- arr drtHead >>> arr (\(Cond c@(CardHead _ _)) -> c) -< e
    cp <- arr (\(CardHead _ cp) -> cp) -< c
    case cp of
      CardEq _ -> writeCardEq -< c
    
writeCardEq :: LA CondHead XmlTree
writeCardEq = proc (CardHead v (CardEq val)) -> do
    returnA <<< mkText -< "|" ++ (formatVar v) ++ "| = " ++ (show val)
    
writeNot :: LA DRT XmlTree
writeNot = text <+> sub
  where
    text = txt "not"
    sub = arr drtSubs >>> arr (!!0) >>> writeExp
      
writePossib :: LA DRT XmlTree
writePossib = text <+> sub
  where
    text = txt "possib"
    sub = arr drtSubs >>> arr (!!0) >>> writeExp
      
writeNecess :: LA DRT XmlTree
writeNecess = text <+> sub
  where
    text = txt "necess"
    sub = arr drtSubs >>> arr (!!0) >>> writeExp
      
writeOr :: LA DRT XmlTree
writeOr = sub1 <+> text <+> sub2
  where
    text = txt " or "
    sub1 = arr drtSubs >>> arr (!!0) >>> writeExp
    sub2 = arr drtSubs >>> arr (!!1) >>> writeExp

writeImpl :: LA DRT XmlTree
writeImpl = sub1 <+> text <+> sub2
  where
    text = txt " impl "
    sub1 = arr drtSubs >>> arr (!!0) >>> writeExp
    sub2 = arr drtSubs >>> arr (!!1) >>> writeExp

writeWhQ :: LA DRT XmlTree
writeWhQ = text <+> sub1 <+> sub2
  where
    text = txt "?"
    sub1 = arr drtSubs >>> arr (!!0) >>> writeExp
    sub2 = arr drtSubs >>> arr (!!1) >>> writeExp
      
writeTimex :: LA DRT XmlTree
writeTimex = txt "timex(" <+> var <+> txt "," <+> name <+> txt ")"
  where
    var  = arr drtHead >>>
           arr (\(Cond (TimexHead v _)) -> v) >>>
           arr formatVar >>>
           mkText
    name = arr drtHead >>>
           arr (\(Cond (TimexHead _ n)) -> n) >>>
           mkText
      
     
-- SDRS writing

writeConstituent :: LA DRT XmlTree
writeConstituent = 
  mkelem "table" [attr "class" (txt "constituent") ] [
    selem "tr" [
      selem "td" [
        writeConstVar
      , txt ": "
      , writeSub
      ]
    ]
  ]
  where
    writeConstVar = arr drtHead >>>
                    arr (\(Const (ConstituentHead v)) -> v) >>>
                    arr formatVar >>>
                    mkText
    writeSub      = arr drtSubs >>> arr (!!0) >>> writeExp
    
writeRelation :: LA Relation XmlTree
writeRelation = proc (Relation n vs) ->
  selem "li" [
    mkText
  ] -< n ++ "(" ++ (concat $ intersperse "," $ map formatVar vs) ++ ")"
