{-# LANGUAGE BangPatterns #-}

module ELCS.DRT.TypeInference (
  TypeError
, inferTypes
, forceInferTypes
, typesAreCompatible
, isFullyTyped
, normalizeDRT

  -- * References
  -- $references
) where

import ELCS.MR
import ELCS.DRT.AST
import qualified ELCS.MR.Type as T
import ELCS.Util

import Data.List
import qualified Data.Map as Map
import Data.Maybe
import Control.Monad
import qualified Control.Monad.State as State
import Control.Exception
--import Debug.Trace

-- | Describes a type error.
type TypeError = String

-- | Adds type annotations to an expression.
--
--   Type inference also respects any previously annotated types, as long as they are complete.
--   If you don’t want this, you should clear type annotations first.
--   Any incomplete types are currently ignored, however.
--
--   Precondition: Variables are assumed to be unique, i.e. no variable of the same name is bound
--   more than once.
inferTypes :: DRT -> Either TypeError DRT
inferTypes e =
  assertMsg ("inferTypes: Vars should be unique in " ++ show e)
            (checkVariablesUnique e) $
  do
    let (eVars, vMap, preConstraints) = generateTypeVars e
    let cs = generateConstraints vMap eVars
    subst <- unify (cs ++ preConstraints)
    let eTyped = refineTypes subst eVars
    return $ eTyped

-- | Infers types, throwing an error if unsuccessful.
forceInferTypes :: DRT -> DRT
forceInferTypes e = case inferTypes e of
                      Right e' -> e'
                      Left err -> throw $ FatalException (err ++ " in " ++ show e)

-- | Returns true if inferTypes succeeds.
isWellTyped :: DRT -> Bool
isWellTyped e = case inferTypes e of
                  Left _ -> False
                  Right _ -> True

-- | Returns true if the two types can be unified.
typesAreCompatible :: T.Type -> T.Type -> Bool
typesAreCompatible t1 t2 = case unifyOne (t1,t2) of
                             Left _ -> False
                             Right _ -> True



-- Generating type variables

-- | Creates a type variable for each MR variable and each expression node.
--   The type variables corresponding to MR variables are stored in the VarMap,
--   and the type variables corresponding to expression nodes are annotated
--   as types on the expression itself.
--   Also emits a list of constraints which relate the newly created type variables
--   to any types that were annotated beforehand.
generateTypeVars :: DRT -> (DRT,VarMap,[Constraint])
generateTypeVars e = State.evalState action initTVarCreator
  where action = do
          vMap <- makeVarMap e
          (e',constraints) <- annotateTypeVars e
          return (e',vMap,constraints)

-- | Generates a type variable for each expression node and annotates it. Also extracts a list
--   of constraints relating to previously annotated types.
annotateTypeVars :: DRT -> TVarCreator (DRT,[Constraint])
annotateTypeVars e = do
  tvar <- getVar
  let oldType = drtType e
  let newType = case oldType of
                  Just t | T.isCompleteType t -> t
                  Just _ | otherwise          -> tvar
                  Nothing                     -> tvar
  subResults <-mapM annotateTypeVars $ drtSubs e
  let subs' = map fst subResults
  let e' = drtSetType newType $ e { drtSubs = subs' }
  return (e',[]) -- The list of constraints is always empty as we only consider complete types,
                 -- which require no extra constraints.

-- | Specifies a mapping of MR variables to type variables.
type VarMap = Map.Map Var T.Type

-- | Generates a type variable for each MR variable and stores them in a map.
makeVarMap :: DRT -> TVarCreator VarMap
makeVarMap e = do
  alist <- forM (vars e) $ \v -> do
             t <- getVar
             return (v,t)
  return $ Map.fromList alist

-- | State monad used for generation of type variables.
type TVarCreator = State.State TVarCreatorState
data TVarCreatorState = TVarCreatorState { varsLeft :: [T.Type] }

-- | A TVarCreator with no used variables.
initTVarCreator :: TVarCreatorState
initTVarCreator = TVarCreatorState { varsLeft = map T.Var [0..] }

-- | Generates a fresh type variable.
getVar :: TVarCreator T.Type
getVar = do
  s <- State.get
  let (v:vs) = varsLeft s
  State.put $ s { varsLeft = vs }
  return v


-- Constraint generation

-- | A constraint asserts the equality of two types.
type Constraint = (T.Type, T.Type)

-- | Walks through the expression and generates appropriate constraints.
--   The expression must have been annotated with type variables.
generateConstraints :: VarMap -> DRT -> [Constraint]
generateConstraints vMap e = let
  -- type of the variable v
  varType v = let mt = Map.lookup v vMap
              in assertMsg ("\n" ++ show v ++ " " ++ (show $ Map.toList vMap) ++ " (" ++ show e ++ ")")
                           (case mt of { Nothing -> False; _ -> True }) $
                 fromJust mt
  -- type of the i-th sub-expression
  subType i = fromJust $ drtType $ (drtSubs e !! i)

  t = fromJust $ drtType e
  cs = case drtHead e of
         Expr (VExpHead v)    -> [ (t, (varType v)) ]
         Expr (AbsHead v)     -> [ (t, (T.Arrow (varType v) (subType 0))) ]
         Expr  AppHead        -> [ ((subType 0), (T.Arrow (subType 1) t)) ]
         Expr (DRSHead vs)    -> [ (t, T.Truth) ] ++
                                 map mkReferentConstraint vs
         Expr (SDRSHead _)    -> [ (t, T.Truth) ]
         Expr  MergeHead      -> [ (t, T.Truth) ] ++
                                 map mkSubTruthConstraint (drtSubs e)
         Expr (AlfaHead _)    -> [ (t, T.Truth) ] ++
                                 map mkSubTruthConstraint (drtSubs e)
         Cond _               -> [ (t, T.Truth) ] ++
                                 map mkReferentConstraint (varsUsedBy e)
         Const (ConstituentHead v) -> [ (t, T.Truth), mkReferentConstraint v ]
  -- Asserts that a given type variable is of type Entity.
  mkReferentConstraint v = ((varType v), T.Entity)
  -- Asserts that a given expression is of type Truth.
  mkSubTruthConstraint s = (fromJust $ drtType s, T.Truth)

  in cs ++ foldr union [] (map (generateConstraints vMap) (drtSubs e))


-- Unification

type Substitution = [(Int,T.Type)]

-- | Unifies the given constraints to yield a substitution which corresponds to the best typing
--   for the expression.
--   This is an adaptation of the nice implementation found at [Cornell 2011].
unify :: [Constraint] -> Either TypeError Substitution
unify ((x,y):cs) = do
  t2 <- unify cs
  t1 <- unifyOne (applySubst t2 x, applySubst t2 y)
  return (t1 ++ t2)
unify [] = Right []

unifyOne :: Constraint -> Either TypeError Substitution
unifyOne c@(s,t) = case c of
  (T.Var v1, T.Var v2) | v1 == v2       -> Right []
                       | otherwise      -> Right [(v1, t)]
  (T.Var v,  _       ) | v `occursIn` t -> Left $ "Circular definition of " ++ show s
                       | otherwise      -> Right [(v, t)]
  (_,        T.Var v ) | v `occursIn` s -> Left $ "Circular definition of " ++ show t
                       | otherwise      -> Right [(v, s)]
  (T.Truth,  T.Truth )                  -> Right []
  (T.Entity, T.Entity)                  -> Right []
  (T.Arrow t1 t2, T.Arrow t1' t2')      -> unify [(t1, t1'), (t2, t2')]
  _                                     -> Left $ "Cannot match " ++ show s ++ " and " ++ show t

applySubst :: Substitution -> T.Type -> T.Type
applySubst subst t = foldr substituteTVar t subst

substituteTVar :: (Int,T.Type) -> T.Type -> T.Type
substituteTVar (v,r) t = case t of
  T.Var v1 | v == v1   -> r
           | otherwise -> t
  T.Arrow t1 t2        -> T.Arrow (substituteTVar (v,r) t1) (substituteTVar (v,r) t2)
  _                    -> t

occursIn :: Int -> T.Type -> Bool
occursIn v t = case t of
  T.Var v1      -> v == v1
  T.Arrow t1 t2 -> occursIn v t1 || occursIn v t2
  _             -> False


-- Applying a substitution to an expression

refineTypes :: Substitution -> DRT -> DRT
refineTypes subst e = let
    t' = applySubst subst (fromJust $ drtType e)
    subs' = map (refineTypes subst) (drtSubs e)
    e' = drtSetType t' $ e { drtSubs = subs' }
  in e'


-- Fully typed expressions

isFullyTyped :: DRT -> Bool
isFullyTyped e = (T.isCompleteType $ fromJust $ drtType e) &&
                 (all isFullyTyped $ drtSubs e)


-- Normalizing expressions, removing extraneous arguments, so they have a full type

normalizer :: DRT
normalizer = mkAbs' ("x",0) (mkDRS' [] [])

-- | Normalises a sentential expression, removing any extraneous arguments it may have.
normalizeDRT :: DRT -> Either (DRT,String) DRT
normalizeDRT e =
  case drtHead e of
    Expr (AbsHead _) ->
      let
        e' = mkApp' e normalizer
      in if isWellTyped e'
           then normalizeDRT $ normalForm $ e'
           else Left (e,"normalizeDRT: Cannot normalise: is not a function in (e->t)")
    _ -> assertMsg ("normalizeDRT: normalised expression is not well-typed\ne: " ++ show e)
                   (isWellTyped e)
                   Right e

{-$references
  * [Cornell 2011] http://www.cs.cornell.edu/courses/cs3110/2011sp/lectures/lec26-type-inference/type-inference.htm#4
-}
