module ELCS.GMB.Document (
-- * Document identifiers
  DocumentId(..)
, mkDocumentId
, parsePartName
, parseDocName
, parseDocId
, partName
, docName
-- * Accessing a GMB instance
, documentList
, docPath
) where

import qualified ELCS.GMB as GMB

import System.Directory
import System.FilePath
import Control.Monad
import Data.List
import Text.Printf
import Text.Regex.TDFA
import Control.Exception


-- | Identifies a single document of the GMB.
data DocumentId =
    DocumentId { partNumber :: Int
                   -- ^ The part number.
               , docNumber :: Int
                   -- ^ The document number.
               }
    deriving (Eq, Show, Ord)
    

-- Helpers for creating and accessing DocumentId
    
mkDocumentId :: Int -> Int -> DocumentId
mkDocumentId p d = DocumentId { partNumber = p, docNumber = d }

parsePartName :: String -> Int
parsePartName = parseGenericName "p"

parseDocName :: String -> Int
parseDocName = parseGenericName "d"

parseGenericName :: (Read a) => String -> String -> a
parseGenericName initial s = let
    re = "^" ++ initial ++ "([[:digit:]]+)$"
    (_,_,_,sm) = (s =~ re) :: (String,String,String,[String])
  in read (sm !! 0)

-- | Parses a string of the form pXXdXXXX  
parseDocId :: String -> Maybe DocumentId
parseDocId s =
  let
    re = "^p([[:digit:]]+)d([[:digit:]]+)$"
    (_,_,_,sm) = (s =~ re) :: (String,String,String,[String])
  in if length sm == 2
       then Just $ mkDocumentId (read (sm !! 0)) (read (sm !! 1))
       else Nothing

partName :: DocumentId -> String
partName d = printf "p%02d" (partNumber d)

docName :: DocumentId -> String
docName d = printf "d%04d" (docNumber d)
  

-- Working with a GMB instance

-- | Returns a list of all documents in the given GMB instance.
documentList :: GMB.Instance -> IO [DocumentId]
documentList (GMB.FileInstance p) = do
  parts <- getSubDirs (p </> "data")
  partDocPaths <- mapM getSubDirs parts
  let docPaths = concat partDocPaths
  let docs = map pathToDoc docPaths
  return docs
documentList (GMB.ZipInstance _p) = undefined
  
pathToDoc :: FilePath -> DocumentId
pathToDoc p =
  assert (length (splitDirectories p) >= 2) $
  let
    dirs = reverse $ splitDirectories p
    doc = dirs !! 0
    part = dirs !! 1
  in mkDocumentId (parsePartName part) (parseDocName doc)
  
-- | Lists all sub-directories of a given path. Helper for documentList.
getSubDirs :: FilePath -> IO [FilePath]
getSubDirs p = do
    s <- getDirectoryContents p
    let s1 = map (\x-> p </> x) $ s \\ [".", ".."]
    s2 <- filterM doesDirectoryExist s1
    return s2

-- | Returns the relative path of the document under the GMB root.
docPath :: DocumentId -> FilePath
docPath d = "data" </> partName d </> docName d
