{-# LANGUAGE Arrows, NoMonomorphismRestriction, FlexibleInstances #-}

module ELCS.GMB.ReadDerivation (
  DRTResult
, readDerivations
, readDerivationsFromFile
, readSentenceMeaningsFromFile
) where

import ELCS.MR
import ELCS.DRT
import ELCS.CCG
import ELCS.GMB as GMB
import ELCS.GMB.Document
import ELCS.Util

import Data.Char
import Data.Maybe
import Text.XML.HXT.Core
import Text.XML.HXT.Arrow.XmlState.RunIOStateArrow
import System.FilePath
import Control.Exception
--import qualified Debug.Trace as Trc


-- We define a state containing a backtrace for processing.

-- User defined state for HXT processing.
data State = State { backtrace :: [String] }
             deriving (Show, Eq)

-- Add an entry to the backtrace, ignoring the arrow input
pushTrace :: String -> a -> State -> State
pushTrace t _ s = s { backtrace = newTrace }
  where newTrace = t : (backtrace s)

-- Remove an entry from the backtrace, ignoring the arrow input
popTrace :: a -> State -> State
popTrace _ s = s { backtrace = newTrace }
  where newTrace = case (backtrace s) of
                     (_:xs) -> xs
                     _      -> []



showTrace :: State -> String
showTrace = show

initialUserState :: State
initialUserState = State { backtrace = [] }

type StateArrow a b = IOStateArrow State a b

runStateArrow :: StateArrow XmlTree c -> IO [c]
runStateArrow = runXIOState (initialState initialUserState)

pushBT :: StateArrow String State
pushBT = proc m -> do
    s <- getUserState -< ()
    returnA <<< setUserState <<< pusher -< (m,s)
  where
    pt :: String -> State -> State
    pt m s = s { backtrace = m : (backtrace s) }

    pusher :: StateArrow (String,State) State
    pusher = proc (m,s) -> do
      returnA <<< arr (uncurry pt) -< (m,s)

popBT :: StateArrow a a
popBT = changeUserState popTrace

getBT :: StateArrow a String
getBT = getUserState >>> arr showTrace

type DRTResult = Maybe DRT


-- Gets the meaning of each sentence
readSentenceMeaningsFromFile :: String -> IO [DRTResult]
readSentenceMeaningsFromFile f = runStateArrow (readDocument hxtOptions f >>>
                                                strictA >>>
                                                extractDerSentenceMeanings)

-- | Reads all derivations in a GMB document.
readDerivations :: GMB.Instance -> DocumentId -> IO [Either [(Derivation, SanityError)] Derivation]
readDerivations gmb doc = do
  let p = docPath doc </> "en.der.xml"
  readDerivationsFromFile gmb p

-- Reads the derivations contained in a given XML file.
readDerivationsFromFile :: GMB.Instance -> FilePath
                        -> IO [Either [(Derivation, SanityError)] Derivation]
readDerivationsFromFile gmb f = do
  ds <- runStateArrow (readGMBFile gmb f >>> getDerivations)
  return $ map saneDerivationOrError ds

-- Arrow for reading in an XML file without any further processing.
readGMBFile :: GMB.Instance -> FilePath -> StateArrow XmlTree XmlTree
readGMBFile gmb f = setTraceLevel 0 >>>
                arrIO0 (GMB.readFile gmb f) >>>
                readFromString hxtOptions

-- | General options used for parsing GMB documents.
hxtOptions :: [SysConfig]
hxtOptions = [ withSubstDTDEntities no
             , withValidate no ]

-- Utility functions

-- Throws an error if application of an arrow would yield zero results.
checkNonEmpty :: String -> StateArrow a b -> StateArrow a b
checkNonEmpty msg a = listA a >>> arr (\x -> if null x then throw $ FatalException msg else x) >>> unlistA

-- Selects all element children with a given name. Checks that there are matching children.
childrenNamed :: String -> StateArrow XmlTree XmlTree
childrenNamed name = checkNonEmpty name (childrenNamed' name)

-- Variant of childrenNamed that may select an empty set of elements
childrenNamed' :: String -> StateArrow XmlTree XmlTree
childrenNamed' name = getChildren >>> isElem >>> hasName name

-- Selects the i-th sub-element (in order of appearance in the file)
selectChild :: Int -> StateArrow XmlTree XmlTree
selectChild i = isElem >>> listA (getChildren >>> isElem) >>>
    proc l -> if (length l) > i
                then arr (!!i) -< l
                else displayError -< ()
  where
    displayError = proc _ -> do
      bt <- getUserState >>. map showTrace -< ()
      errstr <- this -< ("error: selectChild: Index " ++ (show i) ++ " too high" ++ "\n" ++ bt)
      returnA $ traceValue 1 id >>> issueFatal "selectChild" >>> eelem "error" -< errstr

maybeList :: [Maybe a] -> Maybe [a]
maybeList = mapM id


-- Parsing derivations

-- Gets all derivations from the root node of an XML file.
getDerivations :: StateArrow XmlTree Derivation
getDerivations = childrenNamed "xdrs-output" >>>
                 childrenNamed "der" >>>
                 perform (getAttrValue "id" >>> -- Output derivation ID
                          arr (\x -> "Processing derivation " ++ x) >>>
                          traceValue 1 id) >>>
                 getChildren >>> isDerivation >>>
                 getDerivation

-- Parses a derivation tree, given its root (“rule” or “lex”) node.
getDerivation :: StateArrow XmlTree Derivation
getDerivation =
  proc xml -> do
    nodeType <- getNodeType -< xml
    semantics <- arr (makeVariablesUnique . fromJust) <<< ruleMeaning -< xml
    category <- getCategory <<< isElem <<<
                getChildren <<<
                childrenNamed "cat" -< xml
    tokens <- case nodeType of
                LexNode    -> getTokens -< xml -- LexNodes have direct "token" sub-nodes
                RuleNode _ -> getTokens <<< childrenNamed "tokens" -< xml
    children <- listA (getDerivation <<< isDerivation <<< getChildren) -< xml
    returnA <<< strictA -< mkDerivation nodeType (CCGItem tokens category semantics) children

-- Retrieves the node type given a derivation node.
getNodeType :: StateArrow XmlTree NodeType
getNodeType = proc xml -> do
    elemName <- getName -< xml
    typeAttr <- getAttrValue "type" -< xml
    nodeType <- case elemName of
                  "lex"        -> this -< LexNode
                  "unaryrule"  -> this -< RuleNode typeAttr
                  "binaryrule" -> this -< RuleNode typeAttr
                  _            -> arr (\x -> throw $ FatalException ("Invalid derivation node: " ++ x)) -< (elemName)
    returnA -< nodeType

-- Retrieves a syntactic category from a category tag (“atomic”, “backward”, “forward”).
getCategory :: StateArrow XmlTree SynCat
getCategory = proc xml -> do
    elemName <- getName -< xml
    cat <- case elemName of
             "atomic"   -> getAtomic -< xml
             "forward"  -> getComposite ForwardSlash -< xml
             "backward" -> getComposite BackwardSlash -< xml
             _          -> arr (\x -> throw $ FatalException ("Unknown category type: " ++ x)) -< elemName
    returnA -< cat
  where
    getAtomic = proc xml -> do
      val  <- arr concat <<< listA (getText <<< getChildren) -< xml
      feat <- getAttrValue "feature" >>>
              arr (\f -> if f == "" then Nothing else Just f) -< xml
      returnA -< Atom val feat

    getComposite slash = proc xml -> do
      children <- listA (getCategory <<< getChildren) -< xml
      returnA -< slash (children !! 0) (children !! 1)

-- Retrieves the tokens contained inside a <tokens> tag.
getTokens :: StateArrow XmlTree [Token]
getTokens = listA (childrenNamed "token" >>> getChildren >>> getText >>> arr (map toLower))


-- Parsing just the meanings from derivation files

-- Extracts all sentence meanings from a der XML file.
extractDerSentenceMeanings :: StateArrow XmlTree DRTResult
extractDerSentenceMeanings = getChildren >>> isElem >>> sentenceMeanings

-- Given the root of a derivation XML, extracts meaning for every sentence.
sentenceMeanings :: StateArrow XmlTree DRTResult
sentenceMeanings = childrenNamed "der" >>>
                   perform traceDerIndex >>>
                   derRootMeaning
  where
    traceDerIndex  = traceValue 1 traceStr <<< getAttrValue "id"
    traceStr derId = "Processing derivation " ++ derId

-- Given a <der> node, extracts its meaning at the root of the derivation.
derRootMeaning :: StateArrow XmlTree DRTResult
derRootMeaning = getChildren >>> isDerivation >>> ruleMeaning

-- Filter for tags representing derivation nodes.
isDerivation :: StateArrow XmlTree XmlTree
isDerivation = hasName "unaryrule" <+> hasName "binaryrule" <+> hasName "lex"

-- Extracts the semantics of a given rule tag.
ruleMeaning :: StateArrow XmlTree DRTResult
ruleMeaning = childrenNamed "sem" >>>
              selectChild 0 >>>
              getSemantics


-- Parsing DRT expressions

-- Parses an expression.
-- This performs a case match on the tag name to decide what type of expression it’s dealing with.
-- It then delegates the actual parsing to the corresponding function below.
getSemantics :: StateArrow XmlTree DRTResult
getSemantics =
  proc xml -> do
    name <- getName -< xml
    returnA <<< traced doGetSem -< (name, xml)
  where
    traced :: StateArrow (String, XmlTree) DRTResult -> StateArrow (String, XmlTree) DRTResult
    traced a = perform (first pushBT) >>> a >>> perform popBT

    doGetSem :: StateArrow (String, XmlTree) DRTResult
    doGetSem = proc (name, xml) -> case name of
      "lam"      -> getLam               -< xml
      "sdrs"     -> getSdrs              -< xml
      "merge"    -> getMerge             -< xml
      "drs"      -> getDrs               -< xml
      "app"      -> getApp               -< xml
      "var"      -> getVarTag            -< xml
      "alfa"     -> getAlfa              -< xml
      x          -> displayError         -< x

    displayError :: StateArrow String DRTResult
    displayError = proc name -> do
                     msg <- arr ("error: Unknown semantics tag: " ++) -< name
                     bt <- getUserState >>. map showTrace -< ()
                     errstr <- this -< (msg ++ "\n" ++ bt)
                     returnA $ traceValue 1 id >>> constA Nothing -< errstr

getVar :: StateArrow String Var
getVar = arr parseVar

-- Parses a <var> tag.
getVarTag :: StateArrow XmlTree DRTResult
getVarTag = proc xml -> do
    v <- getChildren >>> getText >>> getVar -< xml
    returnA -< Just (mkVExp' v)

-- Parses a <lam> tag.
getLam :: StateArrow XmlTree DRTResult
getLam = proc xml -> do
    varText  <- selectChild 0 >>> getChildren >>> getText     -< xml
    var      <- getVar                                        -< varText
    bodyTree <- selectChild 1                                 -< xml
    body     <- getSemantics                                  -< bodyTree
    returnA -< mkResult var body
  where
    mkResult :: Var -> DRTResult -> DRTResult
    mkResult var (Just body) = Just (mkAbs' var body)
    mkResult _   _           = Nothing

-- Parses an <app> tag.
getApp :: StateArrow XmlTree DRTResult
getApp = proc xml -> do
    sub1  <- selectChild 0 >>> getSemantics -< xml
    sub2  <- selectChild 1 >>> getSemantics -< xml
    returnA -< mkResult sub1 sub2
  where
    mkResult (Just sub1) (Just sub2) = Just (mkApp' sub1 sub2)
    mkResult _           _            = Nothing

-- Parses an <sdrs> tag.
getSdrs :: StateArrow XmlTree DRTResult
getSdrs = proc xml -> do
    consts <- childrenNamed "constituents" >>>
              listA ( checkNonEmpty "constituents" selectConsts ) -< xml
    rels   <- childrenNamed "relations" >>>
              listA (childrenNamed "drel" >>> getRelation)           -< xml
    returnA -< mkResult (maybeList consts) rels
  where
    -- Here’s a clunky workaround: in some documents, constituents are wrapped by a
    -- <sub> tag, in some they aren’t.
    selectConsts = (childrenNamed' "constituent" <+>
                    (childrenNamed' "sub" >>> childrenNamed "constituent")
                   >>> getConstituent)
    mkResult (Just consts) rels = Just (mkSDRS' consts rels)
    mkResult _             _    = Nothing

getConstituent :: StateArrow XmlTree DRTResult
getConstituent = proc c -> do
    v <- getAttrValue0 "label" >>> getVar -< c
    e <- selectChild 0 >>> getSemantics -< c
    returnA -< mkResult v e
  where
    mkResult v (Just e) = Just (mkConstituent' v e)
    mkResult _ _        = Nothing

getRelation :: StateArrow XmlTree Relation
getRelation = proc r -> do
    n  <- getAttrValue0 "sym"                                                -< r
    vs <- listA ((getAttrValue0 "arg1" <+> getAttrValue0 "arg2") >>> getVar) -< r
    returnA -< Relation n vs

-- Parses a <merge> tag.
getMerge :: StateArrow XmlTree DRTResult
getMerge = proc xml -> do
    sub1  <- selectChild 0 >>>
             changeUserState (pushTrace "left") >>>
             getSemantics >>>
             changeUserState popTrace                -< xml
    sub2  <- selectChild 1 >>>
             changeUserState (pushTrace "right") >>>
             getSemantics >>>
             changeUserState popTrace                -< xml
    returnA -< mkResult sub1 sub2
  where
    mkResult (Just sub1) (Just sub2) = Just (mkMerge' sub1 sub2)
    mkResult _           _            = Nothing

-- Parses an <alfa> tag.
-- <alfa type="t">
--   <?>sub1</?>
--   <?>sub2</?>
-- </alfa>
getAlfa :: StateArrow XmlTree DRTResult
getAlfa = proc xml -> do
    t     <- getAttrValue0 "type"                    -< xml
    sub1  <- selectChild 0 >>>
             changeUserState (pushTrace "left") >>>
             getSemantics >>>
             changeUserState popTrace                -< xml
    sub2  <- selectChild 1 >>>
             changeUserState (pushTrace "right") >>>
             getSemantics >>>
             changeUserState popTrace                -< xml
    returnA -< mkResult t sub1 sub2
  where
    mkResult t (Just sub1) (Just sub2) = Just (mkAlfa' t sub1 sub2)
    mkResult _ _           _           = Nothing

-- Parses a <drs> tag.
getDrs :: StateArrow XmlTree DRTResult
getDrs = proc xml -> do
    domain <- childrenNamed "domain" >>>
              listA (childrenNamed' "dr" >>> getDr)     -< xml
    conds  <- childrenNamed "conds" >>>
              listA (annotateConds >>>
                    (traced getCond))                   -< xml
    returnA -< mkResult domain (maybeList conds)
  where
    traced :: StateArrow XmlTree DRTResult -> StateArrow (String,XmlTree) DRTResult
    traced a = perform (first pushBT) >>> arr snd >>> a >>> perform popBT

    -- Adds a numbering string identifier to earch condition
    annotateConds = listA (childrenNamed' "cond") >>>
                    arr (zip condStrs) >>>
                    unlistA
    condStrs = map (\x -> "cond(" ++ show x ++ ")") ([1..]::[Int])

    mkResult domain (Just conds) = Just (mkDRS' domain conds)
    mkResult _      _            = Nothing

getDr :: StateArrow XmlTree Var
getDr = getAttrValue "name" >>> getVar

-- DRS condition parsing
getCond :: StateArrow XmlTree DRTResult
getCond = proc xml -> do
    ctag <- selectChild 0 -< xml
    name <- getName -< ctag
    cond <- case name of
      "pred"  -> getPred      -< ctag
      "rel"   -> getRel       -< ctag
      "named" -> getNamed     -< ctag
      "eq"    -> getEq        -< ctag
      "prop"  -> getProp      -< ctag
      "card"  -> getCard      -< ctag
      "pos"   -> getPos       -< ctag
      "nec"   -> getNec       -< ctag
      "not"   -> getNot       -< ctag
      "imp"   -> getImp       -< ctag
      "or"    -> getOr        -< ctag
      "whq"   -> getWhQ       -< ctag
      "timex" -> getTimex     -< ctag
      x       -> displayError -< x
    returnA -< cond
  where
    displayError = proc name -> do
                     msg <- arr ("error: Unknown DRS condition tag: " ++) -< name
                     bt <- getUserState >>. map showTrace -< ()
                     errstr <- this -< (msg ++ "\n" ++ bt)
                     returnA $ traceValue 1 id >>> constA Nothing -< errstr

getPred :: StateArrow XmlTree DRTResult
getPred = proc xml -> do
    name <- getAttrValue "symbol" -< xml
    arg  <- getVar <<< getAttrValue "arg"  -< xml
    returnA -< Just (mkSimple' name [arg])

getRel :: StateArrow XmlTree DRTResult
getRel = proc xml -> do
    name <- getAttrValue "symbol"            -< xml
    arg1 <- getVar <<< getAttrValue "arg1"   -< xml
    arg2 <- getVar <<< getAttrValue "arg2"   -< xml
    returnA -< Just (mkSimple' name [arg1, arg2])

getNamed :: StateArrow XmlTree DRTResult
getNamed = proc xml -> do
    name <- getAttrValue "symbol"            -< xml
    arg  <- getVar <<< getAttrValue "arg"    -< xml
    returnA -< Just (mkSimple' name [arg])

getEq :: StateArrow XmlTree DRTResult
getEq = proc xml -> do
    arg1 <- getVar <<< getAttrValue "arg1"   -< xml
    arg2 <- getVar <<< getAttrValue "arg2"   -< xml
    returnA -< Just (mkEq' arg1 arg2)

-- Parses a <prop> DRS-condition.
-- <prop argument="v">
--   <indexlist>…</indexlist>
--   <...> Sub-expression </...>
-- </prop>
getProp :: StateArrow XmlTree DRTResult
getProp = proc xml -> do
    v    <- getVar <<< getAttrValue "argument"  -< xml
    body <- getSemantics <<< selectChild 1      -< xml
    returnA -< mkResult v body
  where
    mkResult v (Just body) = Just (mkProp' v body)
    mkResult _ _           = Nothing

-- Parses a <card> DRS-condition.
-- <card type="eq" arg="v" value="x">
--   <indexlist>…</indexlist>
-- </card>
getCard :: StateArrow XmlTree DRTResult
getCard = proc xml -> do
    t    <- getAttrValue "type"                  -< xml
    v    <- getVar <<< getAttrValue "arg"        -< xml
    prop <- case t of
              "eq" -> getCardEq -< xml
              _    -> displayError -< t
    returnA -< mkCond v prop
  where
    mkCond v (Just prop) = Just (mkCard' v prop)
    mkCond _ _           = Nothing

    getCardEq = proc xml -> do
      val  <- getAttrValue "value"                 -< xml
      returnA -< Just (CardEq  (truncate ((read val) :: Double)))

    displayError :: StateArrow String (Maybe CardProp)
    displayError = proc name -> do
      msg  <- arr ("error: Unknown cardinality condition type: " ++) -< name
      bt   <- getBT -< ()
      returnA <<< constA Nothing <<< traceValue 1 id -< (msg ++ "\n" ++ bt)

-- Parses a <pos> DRS-condition.
-- <pos>
--   <indexlist>…</indexlist>
--   <?>sub</?>
-- </pos>
getPos :: StateArrow XmlTree DRTResult
getPos = proc xml -> do
    sub <- getSemantics <<< selectChild 1 -< xml
    returnA -< mkResult sub
  where
    mkResult (Just sub) = Just (mkPossib' sub)
    mkResult _          = Nothing

-- Parses a <nec> DRS-condition
-- <nec>
--   <indexlist>…</indexlist>
--   <?>sub</?>
-- </nec>
getNec :: StateArrow XmlTree DRTResult
getNec = proc xml -> do
    sub <- getSemantics <<< selectChild 1 -< xml
    returnA -< mkResult sub
  where
    mkResult (Just sub) = Just (mkNecess' sub)
    mkResult _          = Nothing

-- Parses a <not> DRS-condition.
-- <not>
--   <indexlist>…</indexlist>
--   <?>sub</?>
-- </not>
getNot :: StateArrow XmlTree DRTResult
getNot = proc xml -> do
    sub <- getSemantics <<< selectChild 1 -< xml
    returnA -< mkResult sub
  where
    mkResult (Just sub) = Just (mkNot' sub)
    mkResult _          = Nothing

-- Parses an <imp> DRS-condition.
-- <imp>
--   <indexlist>…</indexlist>
--   <?>sub1</?>
--   <?>sub2</?>
-- </imp>
getImp :: StateArrow XmlTree DRTResult
getImp = proc xml -> do
    sub1 <- getSemantics <<< selectChild 1 -< xml
    sub2 <- getSemantics <<< selectChild 2 -< xml
    returnA -< mkResult sub1 sub2
  where
    -- Vars need to be moved to the WhQ node because they scope over the second component
    mkResult (Just sub1) (Just sub2) = Just (mkImpl' sub1 sub2)
    mkResult _           _           = Nothing

-- Parses an <or> DRS-condition.
-- <or>
--   <indexlist>…</indexlist>
--   <?>sub1</?>
--   <?>sub2</?>
-- </or>
getOr :: StateArrow XmlTree DRTResult
getOr = proc xml -> do
    sub1 <- getSemantics <<< selectChild 1 -< xml
    sub2 <- getSemantics <<< selectChild 2 -< xml
    returnA -< mkResult sub1 sub2
  where
    mkResult (Just sub1) (Just sub2) = Just (mkOr' sub1 sub2)
    mkResult _           _           = Nothing

-- Parses an <whq> DRS-condition.
-- <whq>
--   <indexlist>…</indexlist>
--   <?>sub1</?>
--   <?>sub2</?>
-- </whq>
getWhQ :: StateArrow XmlTree DRTResult
getWhQ = proc xml -> do
    sub1 <- getSemantics <<< selectChild 1 -< xml
    sub2 <- getSemantics <<< selectChild 2 -< xml
    returnA -< mkResult sub1 sub2
  where
    -- Vars need to be moved to the WhQ node because they scope over the second component
    mkResult (Just sub1) (Just sub2) = Just (mkWhQ' sub1 sub2)
    mkResult _           _           = Nothing

-- Parses a <timex> DRS-condition
-- <timex arg="v">
--   <indexlist>…</indexlist>
--   <date>d</date>
-- </timex>
getTimex :: StateArrow XmlTree DRTResult
getTimex = proc xml -> do
    v <- getVar <<< getAttrValue0 "arg" -< xml
    d <- getText <<< childrenNamed "date" -< xml
    returnA -< Just (mkTimex' v d)
