{-# LANGUAGE FlexibleInstances #-}
{-|
  Contains some basic definitions pertaining to meaning representations.
-}
module ELCS.MR (
-- * Variables
  Var
, varName
, varIndex
, mkVar
, formatVar
, parseVar
, varsByName
, inventVar

-- * Meaning representations
, MR(..)

-- * Building MRs
, MRBuilder
, MRBuilderState
, initMRBuilder
, mrInventVar
) where

import ELCS.MR.Label

import qualified Control.Monad.State.Lazy as S
import Text.Regex.TDFA


-- | A variable is a combination of variable name and index.
--
--   This implementation makes it easy to invent new variables that share a name by increasing the
--   index.
type Var = (String, Int)

-- | Returns the name part of a variable.
varName :: Var -> String
varName (n,_) = n

-- | Returns the index part of a variable.
varIndex :: Var -> Int
varIndex (_,i) = i

-- | Easily create a variable
mkVar :: String -> Int -> Var
mkVar name index = (name, index)

-- | Formats a variable into a string.
formatVar :: Var -> String
formatVar (n,i) = n ++ (show i)

-- | Parses a GMB-style variable string (“x23”, “e42”) into a corresponding variable.
parseVar :: String -> Var
parseVar s = let
    (_,_,_,subms) = (s =~ "^([[:alpha:]]+)([[:digit:]]+)$") :: (String,String,String,[String])
    prefix       = subms !! 0
    suffix       = (read $ subms !! 1) :: Int
  in (prefix, suffix)

-- | The variables in the list that have a certain name
varsByName :: String -> [Var] -> [Var]
varsByName name vs = filter ((== name) . varName) vs

-- | A variable by the given name that is not yet in kv
inventVar :: String -> [Var] -> Var
inventVar name kv = mkVar name (1 + (maxIndex name kv))

-- | The highest index any variable called @name@ has in @kv@.
--   -1 if there is no variable by that name.
maxIndex :: String -> [Var] -> Int
maxIndex name kv = maxOrMinus1 $ map varIndex $ varsByName name kv
  where
    maxOrMinus1 :: [Int] -> Int
    maxOrMinus1 [] = -1
    maxOrMinus1 l = maximum l


-- | Type class for meaning representations.
--
--   A meaning representation is a recursive structure within which variables can occur and be
--   bound. Every node is annotated with an 'MRLabel'.
class Show a => MR a where

  -- | Returns the list of children.
  subExprs :: MR a => a -> [a]

  -- | Returns the label of the expression’s head.
  mrLabel :: MR a => a -> MRLabel
  
  -- | Returns the set of all variables used in the expression.
  vars :: MR a => a -> [Var]
  
  -- | Returns the set of free variables in the given expression.
  freeVars :: MR a => a -> [Var]

    
-- Definition of the MRBuilder monad

-- | MRBuilder is a simple state monad used for creating MRs without using duplicate variables.
type MRBuilder a = S.State MRBuilderState a

-- | State information for MRBuilder.
data MRBuilderState = MRBuilderState { knownVars :: [Var] } deriving (Show,Eq)

-- | Initialises an MRBuilder.
--
--   This function may be passed a list of variables which are presumed as “used” and won’t be
--   issued by 'mrInventVar'.
initMRBuilder :: [Var]          -- ^ The variables already taken.
              -> MRBuilderState -- ^ The resulting state.
initMRBuilder vs = MRBuilderState { knownVars = vs }

-- | Creates a new variable with the given name that is not yet in use.
mrInventVar :: String -> MRBuilder Var
mrInventVar name = do
  s <- S.get
  let kv = knownVars s
  let v = inventVar name kv
  S.put s { knownVars = kv ++ [v] }
  return v

