{-| This module is responsible for parsing command-line options for all sub-commands.
-}
module ELCS.Main.Common.CLOptions (
  parseOptions
, CLOptions(..)
, ExtractOptions(..)
, WDCOptions(..)
, CollateOptions(..)
, SentenceStatsOptions(..)
, GrammarToHtmlOptions(..)
) where

import Options.Applicative

-- | Parses the command line arguments.
parseOptions :: IO CLOptions
parseOptions = execParser clInfo

-- | This sum type contains a separate constructor for each sub-command.
data CLOptions = Extract ExtractOptions
               | WriteDefaultConfig WDCOptions
               | Collate CollateOptions
               | CmdSentenceStats SentenceStatsOptions
               | CmdGrammarToHtml GrammarToHtmlOptions
               
clOptions :: Parser CLOptions
clOptions = subparser
              ( command "Extract" (info extractOptions extractInfo) 
             <> command "WriteDefaultConfig" (info wdcOptions wdcInfo)
             <> command "Collate" (info collateOptions collateInfo)
             <> command "SentenceStats" (info sentenceStatsOptions sentenceStatsInfo)
             <> command "GrammarToHtml" (info grammarToHtmlOptions grammarToHtmlInfo)
              )

clInfo :: ParserInfo CLOptions
clInfo = info ( helper <*> clOptions)
              ( fullDesc
             <> progDesc "Extracts sentence grammars for GMB documents."
             <> header "elcs – a toolkit for semantic CCG induction" )


-- Extract
             
data ExtractOptions = ExtractOptions
  { opt_config :: String
  , opt_gmbpath :: String
  , opt_docs :: String
  , opt_subcorpus :: String
  , opt_output :: String
  , opt_compress :: Bool
  , opt_exclude :: String
  , opt_writeSplits :: Bool
  }
  
extractOptions :: Parser CLOptions
extractOptions = fmap Extract $ ExtractOptions
  <$> strOption
      ( long "config" <> short 'c' <> metavar "PATH" <> value "elcs-extract.config"
     <> help "Path to a config file to use." )
  <*> strOption
      ( long "gmb" <> metavar "PATH" <> value "gmb/"
     <> help "Path to the Groningen Meaning Bank." )
  <*> strOption
      ( long "docs" <> metavar "RANGE" <> value "all"
     <> help "Documents to read." )
  <*> strOption
      ( long "subcorpus" <> metavar "NAME" <> value ""
     <> help "Subcorpus to filter for." )
  <*> strOption
      ( long "output" <> short 'o' <> metavar "PATH" <> value "grammar/"
     <> help "Path to write the extracted sentence grammars to." )
  <*> option
      ( long "compress" <> short 'c' <> metavar "True|False" <> value True
     <> help "Whether to compress output grammars with gzip." )
  <*> strOption
      ( long "exclude" <> metavar "RANGE" <> value ""
     <> help "Documents to exclude" )
  <*> option
      ( long "write-splits" <> metavar "True|False" <> value False
     <> help "Whether to write out an extra file containing all splits that were performed" )
     
extractInfo :: InfoMod CLOptions
extractInfo = progDesc "Extracts grammars for a specified set of sentences."


-- WriteDefaultConfig

data WDCOptions = WDCOptions
 { wdc_output :: String
 }
  
wdcOptions :: Parser CLOptions
wdcOptions = fmap WriteDefaultConfig $ WDCOptions
  <$> strOption
      ( long "output" <> short 'o' <> metavar "PATH" <> value "elcs-extract.config.default"
     <> help "Path to write the default config to." )
     
wdcInfo :: InfoMod CLOptions
wdcInfo = progDesc "Writes a file containing the default configuration, for easy editing."


-- Collate

data CollateOptions = CollateOptions
  { coll_output :: String
  , coll_globs :: [String]
  , coll_inputCodec :: String
  , coll_withStats :: Bool
  , coll_inputWithStats :: Bool
  }
  deriving (Show, Eq)

collateOptions :: Parser CLOptions
collateOptions = fmap Collate $ CollateOptions
  <$> strOption
      ( long "output" <> short 'o' <> metavar "PATH" <> value "out.g.gz"
     <> help "The path for the output grammar." )
  <*> some (argument str (metavar "GLOBS..."))
  <*> strOption
      ( long "input-codec" <> metavar "JSON" <> value "JSON"
     <> help "The format in which the input grammars are." )
  <*> switch
      ( long "with-stats"
     <> help "Whether to produce a stats-annotated grammar." )
  <*> switch
      ( long "input-with-stats"
     <> help "Whether the input files contain statistics." )

collateInfo :: InfoMod CLOptions
collateInfo = progDesc "Collects the contents of multiple grammars."


-- SentenceStats

data SentenceStatsOptions = SentenceStatsOptions
  { sst_input :: String
  , sst_gmb :: String
  } deriving (Show, Eq)
  
sentenceStatsOptions :: Parser CLOptions
sentenceStatsOptions = fmap CmdSentenceStats $ SentenceStatsOptions
  <$> strOption
      ( long "input" <> short 'i' <> metavar "PATH" <> value "."
     <> help "A directory containing sentence grammars." )
  <*> strOption
      ( long "gmb" <> metavar "PATH" <> value "gmb"
     <> help "Path to the GMB." )
     
sentenceStatsInfo :: InfoMod CLOptions
sentenceStatsInfo = progDesc "Collects statistics about sentence grammars."


-- GrammarToHtml

data GrammarToHtmlOptions = GrammarToHtmlOptions
  { gth_input :: String
  } deriving (Show, Eq)
  
grammarToHtmlOptions :: Parser CLOptions
grammarToHtmlOptions = fmap CmdGrammarToHtml $ GrammarToHtmlOptions
  <$> strOption
      ( long "input" <> short 'i' <> metavar "GLOB"
     <> help "Grammar file to convert to HTML." )

grammarToHtmlInfo :: InfoMod CLOptions
grammarToHtmlInfo = progDesc "Saves a grammar as an HTML file."
