{-# LANGUAGE TemplateHaskell #-}

{-| The main module for the Extract command.
-}
module ELCS.Main.Extract (
  mainWith
) where

import ELCS.DRT
import ELCS.CCG
import ELCS.CCG.SplitRecursive
import qualified ELCS.GMB as GMB
import ELCS.GMB.Document
import ELCS.GMB.ReadDerivation
import ELCS.GMB.Metadata
import ELCS.Util
import ELCS.Main.Common.CLOptions
import ELCS.Main.Extract.Config

import Data.List
import Data.List.Split
import Data.Maybe
import System.Directory
import System.FilePath
import System.Timeout
import Control.Exception
import Control.Monad.Reader
import Control.Monad.Error
import Control.DeepSeq
import qualified Control.Concurrent.Thread.Group as ThreadGroup
import Control.Concurrent.Chan
import Text.Printf
import qualified Data.ByteString.Lazy as BS
import qualified Data.Aeson as Aes
import GHC.Conc (numCapabilities)
import Debug.Trace


-- Main plumbing

mainWith :: ExtractOptions -> IO ()
mainWith opts = do
  cfg <- loadExtractorConfig opts
  runReaderT extractMain (ExtractorData opts cfg)
  

-- Concurrent launching of extraction jobs

data Job = SentenceJob { _jobSentence :: (DocumentId, Int)
                       , _jobDerivation :: Derivation
                       , _jobExtractor :: ExtractorData
                       }
         | NoMoreJobs

extractMain :: Extractor ()
extractMain = do
  opts <- asks eOptions
  liftIO $ createDirectoryIfMissing True (opt_output opts)
  let gmb = GMB.initialise (opt_gmbpath opts)
  docs <- getDocs gmb
  jobs <- mkJobs gmb docs
  chan <- liftIO $ newChan
  liftIO $ writeList2Chan chan jobs
  runJobs chan
  
mkJobs :: GMB.Instance -> [DocumentId] -> Extractor [Job]
mkJobs gmb docs = do
  docJobs <- mapM (mkJob gmb) docs
  return (concat docJobs)

mkJob :: GMB.Instance -> DocumentId -> Extractor [Job]
mkJob gmb doc = do
  mders <- liftIO $ readDerivations gmb doc
  let iders = indexSentenceList doc mders
  mjobs <- mapM mkSentenceJob iders
  return $ catMaybes mjobs
  
-- | Handles derivation errors and maybe returns a job.
mkSentenceJob :: ((DocumentId, Int), Either [(Derivation, SanityError)] Derivation)
              -> Extractor (Maybe Job)
mkSentenceJob ((doc, sent), e) = do
  ex <- ask
  case e of
    Left err -> logError (doc,sent) (SEError "Error reading derivation" (show err)) >>
                return Nothing
    Right d  -> return $ Just $ SentenceJob (doc, sent) d ex
    
runJobs :: Chan Job -> Extractor ()
runJobs c = do
  liftIO $ writeChan c NoMoreJobs
  g <- liftIO $ ThreadGroup.new
  forM_ [1..numCapabilities] $ \_-> do
    liftIO $ ThreadGroup.forkIO g $ jobThread c
  liftIO $ ThreadGroup.wait g
    
jobThread :: Chan Job -> IO ()
jobThread c = do
  j <- readChan c
  case j of
    SentenceJob s d ex -> runReaderT (extractAndSave s d) ex  >> jobThread c
    NoMoreJobs -> writeChan c NoMoreJobs -- Put sentinel back for the next job and exit


-- Configuration of the algorithmic components

loadExtractorConfig :: ExtractOptions -> IO Config
loadExtractorConfig opts =
  do
    configJson <- liftIO $ BS.readFile (opt_config opts)
    config <- case (Aes.decode configJson) :: Maybe Config of
                Just c -> return c
                Nothing -> throw $ FatalException "Failed to read config file."
    return config


-- Document extraction

-- | A reader monad that combines command line options and a config file.
type Extractor = ReaderT ExtractorData IO

data ExtractorData =
  ExtractorData
  { eOptions :: ExtractOptions
  , eConfig :: Config
  }
  
-- | Adds document and sentence indices to a list of sentence derivations.
indexSentenceList :: DocumentId -> [a] -> [((DocumentId, Int), a)]
indexSentenceList d xs = zip (zip (repeat d) [1..]) xs

-- | Returns all documents that match the filter.
getDocs :: GMB.Instance -> Extractor [DocumentId]
getDocs gmb = do
  docs <- liftIO $ documentList gmb
  let docsSorted = sort docs
  docsFiltered1 <- filterM docsFilter docsSorted
  mds <- liftIO $ mapM (readMetadata gmb) docsFiltered1
  let mdDocs = zip docsFiltered1 mds
  mdFiltered <- filterM subCorpusFilter mdDocs 
  return (map fst mdFiltered)

  
-- Filtering documents by their IDs

docsFilter :: DocumentId -> Extractor Bool
docsFilter d = do
  isIncluded <- includeFilter d
  isExcluded <- excludeFilter d
  return (isIncluded && not isExcluded)

-- | Returns True if the document is included.
includeFilter :: DocumentId -> Extractor Bool
includeFilter d = do
  opts <- asks eOptions
  let s = opt_docs opts
  if s `elem` ["*","all"] then return True else do
    let components = splitOn "," s
    return $ and $ map (docFilter d) components
    
-- | Returns True if the document is excluded.
excludeFilter :: DocumentId -> Extractor Bool
excludeFilter d = do
  opts <- asks eOptions
  let s = opt_exclude opts
  if null s then return False else do
    let components = splitOn "," s
    return $ or $ map (docFilter d) components
  
docFilter :: DocumentId -> String -> Bool
docFilter d s =
  if ".." `isInfixOf` s then rangeFilter d s else
  case parseDocId s of
    Just docId -> d == docId
    Nothing -> throw $ FatalException ("numFilter: Illegal filter expression \"" ++ s ++ "\"")
  
rangeFilter :: DocumentId -> String -> Bool
rangeFilter d s =
  assert (".." `isInfixOf` s) $
  let
    parts = splitOn ".." s
    [pre,suf] = parts
    parsed@[lo',hi'] = map parseDocId parts
    lo = if null pre then mkDocumentId 0 0 else fromJust lo'
    hi = if null suf then mkDocumentId 99 999 else fromJust hi'
  in if length parts /= 2
       then throw $ FatalException ("rangeFilter: Can only have one occurrence of \"..\"") else
     if all null parts
       then throw $ FatalException ("rangeFilter: All parts empty") else
     if not $ all isJust parsed
       then throw $ FatalException "rangeFilter: Invalid bound" else
     d >= lo && d <= hi


-- Filtering documents by their metadata

subCorpusFilter :: (DocumentId, Metadata) -> Extractor Bool
subCorpusFilter (_, md) =
  asks eOptions >>= \opts ->
  if (opt_subcorpus opts) == "" then return True else
  return $ lookup "subcorpus" md == Just (opt_subcorpus opts)

  
  
-- Sentence extraction
  
type SentenceExtractor = ReaderT SentenceExtractorData (ErrorT SEError IO)

data SentenceExtractorData =
  SentenceExtractorData
  { seOptions :: ExtractOptions 
  , seConfig :: Config 
  , seSentence :: (DocumentId, Int)
  }
  
data SEError = SEError { _seShort :: String, _seLong :: String }
  
instance Error SEError where
  strMsg s =
    if length s > 20
      then SEError (take 19 s ++ "…") s
      else SEError s s
  
extractAndSave :: (DocumentId, Int)
               -> Derivation
               -> Extractor ()
extractAndSave (doc, sent) d =
  do
    opt <- asks eOptions
    cfg <- asks eConfig
    let f = extractAndSaveSentence d
    let rf = runErrorT $ runReaderT f (SentenceExtractorData opt cfg (doc,sent))
    result <- liftIO rf
    case result of
      Left err -> logError (doc,sent) err
      Right s -> return s
    
extractAndSaveSentence :: Derivation -> SentenceExtractor ()
extractAndSaveSentence d = extractTimed d >>= saveExtractionResult
  
extractTimed :: Derivation -> SentenceExtractor ExtractionResult
extractTimed d = do
  opts <- asks seOptions
  cfg <- asks seConfig
  sent <- asks seSentence
  let limit = (cfg_sentenceTimeout cfg)
  let splitCfg = (cfg_alignmentDrivenSplitting cfg)
  (t, mResult) <- liftIO $ time $ timeout limit $ doExtract splitCfg d
  liftIO $ writeFile (opt_output opts </> (sentenceString sent ++ ".TIME")) (show t)
  case mResult of
    Just result -> do
      case result of
        Left err -> throwError (SEError "Error during extraction" err)
        Right r -> do
          let itemCount = resultItemCount r
          liftIO $ traceIO
            (sentenceString sent ++ printf ": Extracted %4d items (%6.0fs)" itemCount t)
          return r
    Nothing -> do
      throwError (SEError "Timeout" "Timeout")
      
-- | Extract catching any FatalExceptions.
doExtract :: Options
          -> Derivation
          -> IO (Either String ExtractionResult)
doExtract opts d = do
  let job = return $!! extract opts d
  mexc <- tryFatalException $ job
  case mexc of
    Left exc -> return (Left (show exc))
    Right noexc -> return noexc

-- Saving

-- | Generates a fitting sentence ID string.
sentenceString :: (DocumentId, Int) -> String
sentenceString (doc, sent) = partName doc ++ docName doc ++ "s" ++ show sent

saveExtractionResult :: ExtractionResult -> SentenceExtractor ()
saveExtractionResult r = do
  opts <- asks seOptions
  saveSentenceGrammar (resultToGrammar r)
  if opt_writeSplits opts
    then saveSplits (resultToSplits r)
    else return ()
  
saveSentenceGrammar :: Grammar DRT -- ^ The grammar to save.
                    -> SentenceExtractor ()
saveSentenceGrammar g = do
  opts <- asks seOptions
  sent <- asks seSentence
  let fName = sentenceString sent ++ ".g" ++
              if opt_compress opts then ".gz" else ""
  liftIO $ saveGrammar (opt_output opts </> fName) g
  
saveSplits :: [(CCGItem DRT, Split DRT)] -> SentenceExtractor ()
saveSplits splits = do
  opts <- asks seOptions
  sent <- asks seSentence
  let fName = sentenceString sent ++ ".splits" ++
              if opt_compress opts then ".gz" else ""
  liftIO $ saveListWith defaultSerialisationOptions (opt_output opts </> fName) splits

logError :: (DocumentId, Int) -> SEError -> Extractor ()
logError (doc, sent) (SEError short long) = do
  opts <- asks eOptions
  let sStr = sentenceString (doc, sent)
  let fName = sStr ++ ".FAIL"
  liftIO $ traceIO (sStr ++ ": Failure (" ++ short ++ ")")
  liftIO $ writeFile (opt_output opts </> fName) long



