module ELCS.Main.SentenceStats (
  mainWith
) where

import ELCS.Main.Common.CLOptions
import ELCS.CCG
import ELCS.DRT
import qualified ELCS.GMB as GMB
import ELCS.GMB.ReadDerivation
import ELCS.GMB.Document
import ELCS.Util

import Data.List
import Text.Printf
import System.IO
import System.FilePath
import System.FilePath.Glob
import Control.Exception
import Control.Monad
import Control.Monad.State
import Control.Applicative
import Debug.Trace

mainWith :: SentenceStatsOptions -> IO ()
mainWith opts = do
  gFiles <- globGrammars (sst_input opts)
  evalStateT (processGrammars gFiles) (initialState opts)
  
  
data SentenceStats = SentenceStats
  { sentIndex :: Int
  , sentId :: String
  , sentWordCount :: Int
  , sentItemCount :: Int
  , sentExtractionTime :: Double
  , sentMrSize :: Int
  , sentAlignedNodes :: Int
  } deriving (Eq, Show, Ord)
  
data WordStats = WordStats
  { wordIndex :: Int
  , wordId :: String
  , wordToken :: String
  , wordSmallestItemSize :: Int
  } deriving (Eq, Show, Ord)

sentenceStatsFields :: [(String, SentenceStats -> String)]
sentenceStatsFields =
  [ ("index",          show . sentIndex)
  , ("id",             sentId)
  , ("wordCount",      show . sentWordCount)
  , ("itemCount",      show . sentItemCount)
  , ("extractionTime", printf "%.2f" . sentExtractionTime)
  , ("mrSize",         show . sentMrSize)
  , ("alignedNodes",   show . sentAlignedNodes)
  ]
  
wordStatsFields :: [(String, WordStats -> String)]
wordStatsFields =
  [ ("index",            show . wordIndex)
  , ("id",               wordId)
  , ("token",            wordToken)
  , ("smallestItemSize", show . wordSmallestItemSize)
  ]

extractSentenceStats :: (DocumentId, Int) -> Grammar DRT -> Derivation -> StatsWriter SentenceStats
extractSentenceStats (doc, sent) g d = do
  index <- incrCurrentSentenceIndex
  return $ SentenceStats
    { sentIndex = index
    , sentId = makeSentenceId doc sent
    , sentWordCount = length $ derivTokens d
    , sentItemCount = length g
    , sentExtractionTime = 0.0 -- todo
    , sentMrSize = 0 -- todo
    , sentAlignedNodes = 0 -- todo
    }

extractWordStats :: (DocumentId, Int) -> Grammar DRT -> Derivation -> StatsWriter [WordStats]
extractWordStats (doc, sent) g d = do
  let indices = [0..length (derivTokens d)-1]
  let smallestItemSizes = map (findSmallestItemSize g) (derivTokens d)
  forM indices $ \index -> do
    windex <- incrCurrentWordIndex
    return $ WordStats
      { wordIndex = windex
      , wordId = makeWordId doc sent index
      , wordToken = derivTokens d !! index
      , wordSmallestItemSize = smallestItemSizes !! index
      }
      
findSmallestItemSize :: Grammar DRT -> String -> Int
findSmallestItemSize g w = minimum $ map itemLength g
  where itemLength item = if w `elem` ccgTokens item
                            then length (ccgTokens item)
                            else maxBound

  
-- StatsWriter monad

type StatsWriter = StateT StatsWriterState IO

data StatsWriterState = StatsWriterState
  { stateCurrentSentenceIndex :: Int
  , stateCurrentWordIndex :: Int
  , stateGmb :: GMB.Instance
  , stateDerivationCache :: Maybe (DocumentId, [Either [(Derivation, SanityError)] Derivation])
  } deriving (Show)
  
initialState :: SentenceStatsOptions -> StatsWriterState
initialState opts = StatsWriterState 0 0 (GMB.initialise (sst_gmb opts)) Nothing

incrCurrentSentenceIndex :: StatsWriter Int
incrCurrentSentenceIndex = do
  i <- gets stateCurrentSentenceIndex
  modify (\s -> s { stateCurrentSentenceIndex = i+1 })
  return i

incrCurrentWordIndex :: StatsWriter Int
incrCurrentWordIndex = do
  i <- gets stateCurrentWordIndex
  modify (\s -> s { stateCurrentWordIndex = i+1 })
  return i

  
globGrammars :: FilePath -> IO [FilePath]
globGrammars p = sort <$> concat <$> (mapM glob [p </> "*.g", p </> "*.g.gz"])

processGrammars :: [FilePath] -> StatsWriter ()
processGrammars fs = do
  withStatsFiles $ \(sentFile, wordFile) ->
    forM_ fs $ \f -> do
      liftIO $ traceIO f
      let (doc, sent) = parseFileName f
      g <- liftIO $ loadGrammarDRT f
      d <- getDerivation (doc, sent)
      sentStats <- extractSentenceStats (doc, sent) g d
      wordStats <- extractWordStats (doc, sent) g d
      liftIO $ writeSentenceStatsRecord sentFile sentStats
      liftIO $ writeWordStatsRecords wordFile wordStats
      liftIO $ hFlush sentFile
      liftIO $ hFlush wordFile

getDerivation :: (DocumentId, Int) -> StatsWriter Derivation
getDerivation (doc, sent) = do
  cache <- gets stateDerivationCache
  d <- case cache of
         Just (cacheId, cacheDeriv) ->
           if doc == cacheId
             then return $ cacheDeriv !! sent
             else readSentenceDerivation doc sent
         Nothing -> readSentenceDerivation doc sent
  case d of
    Right der -> return der
    Left e -> throw $ FatalException ("Failed to read sentence derivation" ++ show e)

readSentenceDerivation :: DocumentId -> Int -> StatsWriter (Either [(Derivation, SanityError)] Derivation)
readSentenceDerivation doc sent = do
  gmb <- gets stateGmb
  ds <- liftIO $ readDerivations gmb doc
  modify (\s -> s { stateDerivationCache = Just (doc, ds) })
  return $ ds !! sent
        
withStatsFiles :: ((Handle, Handle) -> StatsWriter ()) -> StatsWriter ()
withStatsFiles a = do
  -- todo exception handling could be nicer here (i.e., existent)
  sentFile <- liftIO $ openFile "sentence-stats.tsv" WriteMode
  wordFile <- liftIO $ openFile "word-stats.tsv" WriteMode
  liftIO $ writeSentenceStatsHeader sentFile
  liftIO $ writeWordStatsHeader wordFile
  a (sentFile, wordFile)
  liftIO $ hClose wordFile
  liftIO $ hClose sentFile
  
parseFileName :: FilePath -> (DocumentId, Int)
parseFileName f =
  let
    docStr = (take 8 $ takeFileName f)           -- Parse only the pXXdXXXX part.
    sent = (-1 +) $ read $ drop 9 $ dropExtensions $ takeFileName f :: Int -- Parse the sentence index
    doc = case parseDocId docStr of
            Just d -> d
            Nothing -> throw $ FatalException ("Failed to parse document ID " ++ docStr)
  in (doc, sent)
  
makeSentenceId :: DocumentId -> Int -> String
makeSentenceId doc sent = printf "p%02dd%04ds%02d" (partNumber doc) (docNumber doc) sent

makeWordId :: DocumentId -> Int -> Int -> String
makeWordId doc sent word = makeSentenceId doc sent ++ printf "w%02d" word

writeSentenceStatsHeader :: Handle -> IO ()
writeSentenceStatsHeader = writeTSVHeader (map fst sentenceStatsFields)

writeSentenceStatsRecord :: Handle -> SentenceStats -> IO ()
writeSentenceStatsRecord = writeTSVRecord (map snd sentenceStatsFields)

writeWordStatsHeader :: Handle -> IO ()
writeWordStatsHeader = writeTSVHeader (map fst wordStatsFields)

writeWordStatsRecords :: Handle -> [WordStats] -> IO ()
writeWordStatsRecords h rs = forM_ rs (writeTSVRecord (map snd wordStatsFields) h)

writeTSVHeader :: [String] -> Handle -> IO ()
writeTSVHeader fields h = do
  let line = concat $ intersperse "\t" fields
  hPutStrLn h line

writeTSVRecord :: [(a -> String)] -> Handle -> a -> IO ()
writeTSVRecord fs h r = do
  let line = concat $ intersperse "\t" $ map ($ r) fs
  hPutStrLn h line
